#
# Copyright (c) 2018, Salesforce, Inc.
#                     The Board of Trustees of the Leland Stanford Junior University
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import csv
import glob
import hashlib
import io
import json
import logging
import os
import re
import unicodedata
import xml.etree.ElementTree as ET
from typing import Iterable

import torch

from ..data_utils.example import Example, NumericalizedExamples
from .base_dataset import Dataset, Split, interleave_keys

logger = logging.getLogger(__name__)


def make_example_id(dataset, example_id):
    return dataset.name + '/' + str(example_id)


# sort_key funcs
def context_answer_len(ex: NumericalizedExamples):
    return interleave_keys(ex.context.length, ex.answer.length)


def context_question_len(ex: NumericalizedExamples):
    return ex.context.length  # question is already appended to context


def input_then_output_len(ex: NumericalizedExamples):
    """
    sort by input length, break ties by output length
    """
    return (context_question_len(ex), answer_len(ex))


def answer_len(ex: NumericalizedExamples):
    return ex.answer.length


def id_value(ex):
    id_ = ex.example_id.rsplit('/', 1)
    id_ = id_[0] if len(id_) == 1 else id_[1]
    return id_


# batch_size functions; batch size is calculated after pad tokens are added
def input_tokens_fn(batch: Iterable[NumericalizedExamples]):
    return max([context_question_len(e) for e in batch]) * len(batch)


def all_tokens_fn(batch: Iterable[NumericalizedExamples]):
    return (max([context_question_len(e) for e in batch]) + max([answer_len(e) for e in batch])) * len(batch)


def default_batch_fn(batch: Iterable[NumericalizedExamples]):
    return len(batch)


class CQA(Dataset):
    def __init__(self, examples, sort_key_fn=input_then_output_len, batch_size_fn=all_tokens_fn, groups=None, **kwargs):
        self.sort_key_fn = sort_key_fn
        self.batch_size_fn = batch_size_fn
        self.groups = groups
        super().__init__(examples, **kwargs)


class IMDb(CQA):
    urls = ['http://ai.stanford.edu/~amaas/data/sentiment/aclImdb_v1.tar.gz']
    name = 'imdb'
    dirname = 'aclImdb'

    def __init__(self, path, subsample=None, lower=False, cached_path=None, skip_cache=False, **kwargs):
        examples = []
        labels = {'neg': 'negative', 'pos': 'positive'}
        question = 'Is this review negative or positive?'

        cache_name = os.path.join(cached_path, os.path.basename(path), str(subsample))
        if os.path.exists(cache_name) and not skip_cache:
            logger.info(f'Loading cached data from {cache_name}')
            examples = torch.load(cache_name)
        else:
            for label in ['pos', 'neg']:
                for fname in glob.iglob(os.path.join(path, label, '*.txt')):
                    with open(fname, 'r') as f:
                        context = f.readline()
                    answer = labels[label]
                    examples.append(
                        Example.from_raw(make_example_id(self, len(examples)), context, question, answer, lower=lower)
                    )
                    if subsample is not None and len(examples) > subsample:
                        break
            os.makedirs(os.path.dirname(cache_name), exist_ok=True)
            logger.info(f'Caching data to {cache_name}')
            torch.save(examples, cache_name)
        super().__init__(examples, **kwargs)

    @classmethod
    def splits(cls, root='.data', train='train', validation=None, test='test', **kwargs):
        assert validation is None
        path = cls.download(root)

        train_data = None if train is None else cls(os.path.join(path, f'{train}'), **kwargs)
        test_data = None if test is None else cls(os.path.join(path, f'{test}'), **kwargs)

        aux_data = None
        do_curriculum = kwargs.get('curriculum', False)
        if do_curriculum:
            kwargs.pop('curriculum')
            aux_data = cls(os.path.join(path, 'aux' + '.tsv'), **kwargs)

        return Split(
            train=None if train is None else train_data,
            eval=None,
            test=None if test is None else test_data,
            aux=None if do_curriculum is None else aux_data,
        )


class SST(CQA):
    urls = [
        'https://raw.githubusercontent.com/openai/generating-reviews-discovering-sentiment/master/data/train_binary_sent.csv',
        'https://raw.githubusercontent.com/openai/generating-reviews-discovering-sentiment/master/data/dev_binary_sent.csv',
        'https://raw.githubusercontent.com/openai/generating-reviews-discovering-sentiment/master/data/test_binary_sent.csv',
    ]
    name = 'sst'
    dirname = ''

    def __init__(self, path, subsample=None, lower=False, cached_path=None, skip_cache=False, **kwargs):
        cache_name = os.path.join(cached_path, os.path.basename(path), str(subsample))

        examples = []
        if os.path.exists(cache_name) and not skip_cache:
            logger.info(f'Loading cached data from {cache_name}')
            examples = torch.load(cache_name)
        else:
            labels = ['negative', 'positive']
            question = 'Is this review ' + labels[0] + ' or ' + labels[1] + '?'

            with io.open(os.path.expanduser(path), encoding='utf8') as f:
                next(f)
                for line in f:
                    parsed = list(csv.reader([line.rstrip('\n')]))[0]
                    context = parsed[-1]
                    answer = labels[int(parsed[0])]
                    examples.append(
                        Example.from_raw(make_example_id(self, len(examples)), context, question, answer, lower=lower)
                    )

                    if subsample is not None and len(examples) > subsample:
                        break

            os.makedirs(os.path.dirname(cache_name), exist_ok=True)
            logger.info(f'Caching data to {cache_name}')
            torch.save(examples, cache_name)

        self.examples = examples
        super().__init__(examples, **kwargs)

    @classmethod
    def splits(cls, root='.data', train='train', validation='dev', test='test', **kwargs):
        path = cls.download(root)
        postfix = '_binary_sent.csv'

        train_data = None if train is None else cls(os.path.join(path, f'{train}{postfix}'), **kwargs)
        validation_data = None if validation is None else cls(os.path.join(path, f'{validation}{postfix}'), **kwargs)
        test_data = None if test is None else cls(os.path.join(path, f'{test}{postfix}'), **kwargs)

        aux_data = None
        do_curriculum = kwargs.get('curriculum', False)
        if do_curriculum:
            kwargs.pop('curriculum')
            aux_data = cls(os.path.join(path, f'aux{postfix}'), **kwargs)

        return Split(
            train=None if train is None else train_data,
            eval=None if validation is None else validation_data,
            test=None if test is None else test_data,
            aux=None if do_curriculum is None else aux_data,
        )


class TranslationDataset(CQA):
    def __init__(self, path, exts, subsample=None, lower=False, cached_path=None, skip_cache=False, **kwargs):
        """Create a TranslationDataset given paths and fields.

        Arguments:
            path: Common prefix of paths to the data files for both languages.
            exts: A tuple containing the extension to path for each language.
            fields$: fields for handling all columns
            Remaining keyword arguments: Passed to the constructor of Dataset.
        """
        cache_name = os.path.join(cached_path, os.path.basename(path), str(subsample))

        if os.path.exists(cache_name) and not skip_cache:
            logger.info(f'Loading cached data from {cache_name}')
            examples = torch.load(cache_name)
        else:
            langs = {
                '.de': 'German',
                '.en': 'English',
                '.fr': 'French',
                '.ar': 'Arabic',
                '.cs': 'Czech',
                '.tt': 'ThingTalk',
                '.fa': 'Farsi',
            }
            source, target = langs[exts[0]], langs[exts[1]]
            src_path, trg_path = tuple(os.path.expanduser(path + x) for x in exts)
            question = f'Translate from {source} to {target}'

            examples = []
            with open(src_path) as src_file, open(trg_path) as trg_file:
                for src_line, trg_line in zip(src_file, trg_file):
                    src_line, trg_line = src_line.strip(), trg_line.strip()
                    if src_line != '' and trg_line != '':
                        context = src_line
                        answer = trg_line
                        examples.append(
                            Example.from_raw(make_example_id(self, len(examples)), context, question, answer, lower=lower)
                        )
                        if subsample is not None and len(examples) >= subsample:
                            break

            os.makedirs(os.path.dirname(cache_name), exist_ok=True)
            logger.info(f'Caching data to {cache_name}')
            torch.save(examples, cache_name)
        super().__init__(examples, **kwargs)

    @classmethod
    def splits(cls, exts, root='.data', train='train', validation='val', test='test', **kwargs):
        """Create dataset objects for splits of a TranslationDataset.

        Arguments:

            root: Root dataset storage directory. Default is '.data'.
            exts: A tuple containing the extension to path for each language.
            fields: A tuple containing the fields that will be used for data
                in each language.
            train: The prefix of the train data. Default: 'train'.
            validation: The prefix of the validation data. Default: 'val'.
            test: The prefix of the test data. Default: 'test'.
            Remaining keyword arguments: Passed to the splits method of
                Dataset.
        """
        path = cls.download(root)

        train_data = None if train is None else cls(os.path.join(path, train), exts, **kwargs)
        val_data = None if validation is None else cls(os.path.join(path, validation), exts, **kwargs)
        test_data = None if test is None else cls(os.path.join(path, test), exts, **kwargs)

        aux_data = None
        do_curriculum = kwargs.get('curriculum', False)
        if do_curriculum:
            kwargs.pop('curriculum')
            aux_data = cls(os.path.join(path, 'aux'), exts, **kwargs)

        return Split(
            train=None if train is None else train_data,
            eval=None if validation is None else val_data,
            test=None if test is None else test_data,
            aux=None if do_curriculum is None else aux_data,
        )


class Multi30k(TranslationDataset, CQA):
    urls = [
        'http://www.quest.dcs.shef.ac.uk/wmt16_files_mmt/training.tar.gz',
        'http://www.quest.dcs.shef.ac.uk/wmt16_files_mmt/validation.tar.gz',
        'http://www.quest.dcs.shef.ac.uk/wmt17_files_mmt/mmt_task1_test2016.tar.gz',
    ]
    name = 'multi30k'
    dirname = ''


class IWSLT(TranslationDataset, CQA):
    base_url = 'https://wit3.fbk.eu/archive/2016-01//texts/{}/{}/{}.tgz'
    name = 'iwslt'
    base_dirname = '{}-{}'

    @classmethod
    def splits(cls, exts, root='.data', train='train', validation='IWSLT16.TED.tst2013', test='IWSLT16.TED.tst2014', **kwargs):
        """Create dataset objects for splits of the IWSLT dataset.

        Arguments:

            root: Root dataset storage directory. Default is '.data'.
            exts: A tuple containing the extension to path for each language.
            fields: A tuple containing the fields that will be used for data
                in each language.
            train: The prefix of the train data. Default: 'train'.
            validation: The prefix of the validation data. Default: 'val'.
            test: The prefix of the test data. Default: 'test'.
            Remaining keyword arguments: Passed to the splits method of
                Dataset.
        """
        cls.dirname = cls.base_dirname.format(exts[0][1:], exts[1][1:])
        cls.urls = [cls.base_url.format(exts[0][1:], exts[1][1:], cls.dirname)]
        check = os.path.join(root, cls.name, cls.dirname)
        path = cls.download(root, check=check)

        if train is not None:
            train = '.'.join([train, cls.dirname])
        if validation is not None:
            validation = '.'.join([validation, cls.dirname])
        if test is not None:
            test = '.'.join([test, cls.dirname])

        if not os.path.exists(os.path.join(path, '.'.join(['train', cls.dirname])) + exts[0]):
            cls.clean(path)

        train_data = None if train is None else cls(os.path.join(path, train), exts, **kwargs)
        val_data = None if validation is None else cls(os.path.join(path, validation), exts, **kwargs)
        test_data = None if test is None else cls(os.path.join(path, test), exts, **kwargs)

        aux_data = None
        do_curriculum = kwargs.get('curriculum', False)
        if do_curriculum:
            kwargs.pop('curriculum')
            aux = '.'.join(['aux', cls.dirname])
            aux_data = cls(os.path.join(path, aux), exts, **kwargs)

        return Split(
            train=None if train is None else train_data,
            eval=None if validation is None else val_data,
            test=None if test is None else test_data,
            aux=None if do_curriculum is None else aux_data,
        )

    @staticmethod
    def clean(path):
        for f_xml in glob.iglob(os.path.join(path, '*.xml')):
            print(f_xml)
            f_txt = os.path.splitext(f_xml)[0]
            with io.open(f_txt, mode='w', encoding='utf-8') as fd_txt:
                root = ET.parse(f_xml).getroot()[0]
                for doc in root.findall('doc'):
                    for e in doc.findall('seg'):
                        fd_txt.write(e.text.strip() + '\n')

        xml_tags = ['<url', '<keywords', '<talkid', '<description', '<reviewer', '<translator', '<title', '<speaker']
        for f_orig in glob.iglob(os.path.join(path, 'train.tags*')):
            print(f_orig)
            f_txt = f_orig.replace('.tags', '')
            with io.open(f_txt, mode='w', encoding='utf-8') as fd_txt, io.open(f_orig, mode='r', encoding='utf-8') as fd_orig:
                for l_ in fd_orig:
                    if not any(tag in l_ for tag in xml_tags):
                        fd_txt.write(l_.strip() + '\n')


class SQuAD(CQA):
    urls = [
        'https://rajpurkar.github.io/SQuAD-explorer/dataset/train-v1.1.json',
        'https://rajpurkar.github.io/SQuAD-explorer/dataset/dev-v1.1.json',
        'https://rajpurkar.github.io/SQuAD-explorer/dataset/train-v2.0.json',
        'https://rajpurkar.github.io/SQuAD-explorer/dataset/dev-v2.0.json',
    ]
    name = 'squad'
    dirname = ''

    def __init__(self, path, subsample=None, lower=False, cached_path=None, skip_cache=False, **kwargs):
        cache_name = os.path.join(cached_path, os.path.basename(path), str(subsample))

        examples, all_answers, q_ids = [], [], []
        if os.path.exists(cache_name) and not skip_cache:
            logger.info(f'Loading cached data from {cache_name}')
            examples, all_answers, q_ids = torch.load(cache_name)
        else:
            with open(os.path.expanduser(path)) as f:
                squad = json.load(f)['data']
                for document in squad:
                    paragraphs = document['paragraphs']
                    for paragraph in paragraphs:
                        context = paragraph['context']
                        qas = paragraph['qas']
                        for qa in qas:
                            question = ' '.join(qa['question'].split())
                            q_ids.append(qa['id'])
                            if len(qa['answers']) == 0:
                                answer = 'unanswerable'
                                all_answers.append(['unanswerable'])
                                context = ' '.join(context.split())
                                ex = Example.from_raw(make_example_id(self, qa['id']), context, question, answer, lower=lower)
                            else:
                                answer = qa['answers'][0]['text']
                                all_answers.append([a['text'] for a in qa['answers']])
                                answer_start = qa['answers'][0]['answer_start']
                                answer_end = answer_start + len(answer)
                                context_before_answer = context[:answer_start]
                                context_after_answer = context[answer_end:]
                                BEGIN = 'beginanswer '
                                END = ' endanswer'

                                tagged_context = context_before_answer + BEGIN + answer + END + context_after_answer
                                tagged_context = tagged_context.split()

                                tokenized_answer = answer.split()
                                for xi, x in enumerate(tagged_context):
                                    if BEGIN in x:
                                        answer_start = xi + 1
                                        tagged_context[xi] = x.replace(BEGIN, '')
                                    if END in x:
                                        answer_end = xi
                                        tagged_context[xi] = x.replace(END, '')
                                new_context = []
                                original_answer_start = answer_start
                                original_answer_end = answer_end
                                indexed_with_spaces = tagged_context[answer_start:answer_end]
                                if len(indexed_with_spaces) != len(tokenized_answer):
                                    import pdb

                                    pdb.set_trace()

                                # remove spaces
                                for xi, x in enumerate(tagged_context):
                                    if len(x.strip()) == 0:
                                        if xi <= original_answer_start:
                                            answer_start -= 1
                                        if xi < original_answer_end:
                                            answer_end -= 1
                                    else:
                                        new_context.append(x)
                                tagged_context = new_context
                                tokenized_answer = [x for x in tokenized_answer if len(x.strip()) > 0]
                                if len(tagged_context[answer_start:answer_end]) != len(tokenized_answer):
                                    import pdb

                                    pdb.set_trace()
                                context_spans = list(range(answer_start, answer_end))
                                indexed_answer = tagged_context[context_spans[0] : context_spans[-1] + 1]
                                if len(indexed_answer) != len(tokenized_answer):
                                    import pdb

                                    pdb.set_trace()
                                context_spans += [len(tagged_context)]
                                for context_idx, answer_word in zip(context_spans, ex.answer):
                                    if context_idx == len(tagged_context):
                                        continue
                                    if tagged_context[context_idx] != answer_word:
                                        import pdb

                                        pdb.set_trace()

                                ex = Example.from_raw(
                                    make_example_id(self, qa['id']),
                                    ' '.join(tagged_context),
                                    question,
                                    ' '.join(tokenized_answer),
                                    lower=lower,
                                )

                            examples.append(ex)
                            if subsample is not None and len(examples) > subsample:
                                break
                        if subsample is not None and len(examples) > subsample:
                            break
                    if subsample is not None and len(examples) > subsample:
                        break

            os.makedirs(os.path.dirname(cache_name), exist_ok=True)
            logger.info(f'Caching data to {cache_name}')
            torch.save((examples, all_answers, q_ids), cache_name)

        super(SQuAD, self).__init__(examples, **kwargs)
        self.all_answers = all_answers
        self.q_ids = q_ids

    @classmethod
    def splits(cls, root='.data', description='squad1.1', train='train', validation='dev', test=None, **kwargs):
        """Create dataset objects for splits of the SQuAD dataset.
        Arguments:
            root: directory containing SQuAD data
            field: field for handling all columns
            train: The prefix of the train data. Default: 'train'.
            validation: The prefix of the validation data. Default: 'val'.
            Remaining keyword arguments: Passed to the splits method of
                Dataset.
        """
        assert test is None
        path = cls.download(root)

        extension = 'v2.0.json' if '2.0' in description else 'v1.1.json'

        train = '-'.join([train, extension]) if train is not None else None
        validation = '-'.join([validation, extension]) if validation is not None else None

        train_data = None if train is None else cls(os.path.join(path, train), **kwargs)
        validation_data = None if validation is None else cls(os.path.join(path, validation), **kwargs)

        aux_data = None
        do_curriculum = kwargs.get('curriculum', False)
        if do_curriculum:
            kwargs.pop('curriculum')
            aux = '-'.join(['aux', extension])
            aux_data = cls(os.path.join(path, aux), **kwargs)

        return Split(
            train=None if train is None else train_data,
            eval=None if validation is None else validation_data,
            test=None,
            aux=None if do_curriculum is None else aux_data,
        )


# https://github.com/abisee/cnn-dailymail/blob/8eace60f306dcbab30d1f1d715e379f07a3782db/make_datafiles.py
dm_single_close_quote = u'\u2019'
dm_double_close_quote = u'\u201d'
# acceptable ways to end a sentence
END_TOKENS = ['.', '!', '?', '...', "'", "`", '"', dm_single_close_quote, dm_double_close_quote, ")"]


def fix_missing_period(line):
    """Adds a period to a line that is missing a period"""
    if "@highlight" in line:
        return line
    if line == "":
        return line
    if line[-1] in END_TOKENS:
        return line
    return line + "."


class Summarization(CQA):
    def __init__(self, path, one_answer=True, subsample=None, lower=False, cached_path=None, skip_cache=False, **kwargs):
        cache_name = os.path.join(cached_path, os.path.basename(path), str(subsample))

        examples = []
        if os.path.exists(cache_name) and not skip_cache:
            logger.info(f'Loading cached data from {cache_name}')
            examples = torch.load(cache_name)
        else:
            with open(os.path.expanduser(path)) as f:
                lines = f.readlines()
                for line in lines:
                    ex = json.loads(line)
                    context, question, answer = ex['context'], ex['question'], ex['answer']
                    examples.append(
                        Example.from_raw(make_example_id(self, len(examples)), context, question, answer, lower=lower)
                    )
                    if subsample is not None and len(examples) >= subsample:
                        break
            os.makedirs(os.path.dirname(cache_name), exist_ok=True)
            logger.info(f'Caching data to {cache_name}')
            torch.save(examples, cache_name)

        super(Summarization, self).__init__(examples, **kwargs)

    @classmethod
    def cache_splits(cls, path):

        splits = ['training', 'validation', 'test']
        for split in splits:
            missing_stories, collected_stories = 0, 0
            split_file_name = os.path.join(path, f'{split}.jsonl')
            if os.path.exists(split_file_name):
                continue
            with open(split_file_name, 'w') as split_file:
                url_file_name = os.path.join(path, f'{cls.name}_wayback_{split}_urls.txt')
                with open(url_file_name) as url_file:
                    for url in url_file:
                        story_file_name = os.path.join(
                            path, 'stories', f"{hashlib.sha1(url.strip().encode('utf-8')).hexdigest()}.story"
                        )
                        try:
                            story_file = open(story_file_name)
                        except EnvironmentError as e:
                            missing_stories += 1
                            logger.warning(e)
                            if os.path.exists(split_file_name):
                                os.remove(split_file_name)
                        else:
                            with story_file:
                                article, highlight = [], []
                                is_highlight = False
                                for line in story_file:
                                    line = line.strip()
                                    if line == "":
                                        continue
                                    line = fix_missing_period(line)
                                    if line.startswith("@highlight"):
                                        is_highlight = True
                                    elif "@highlight" in line:
                                        raise ValueError()
                                    elif is_highlight:
                                        highlight.append(line)
                                    else:
                                        article.append(line)
                                example = {
                                    'context': unicodedata.normalize('NFKC', ' '.join(article)),
                                    'answer': unicodedata.normalize('NFKC', ' '.join(highlight)),
                                    'question': 'What is the summary?',
                                }
                                split_file.write(json.dumps(example) + '\n')
                                collected_stories += 1
                                if collected_stories % 1000 == 0:
                                    logger.debug(example)
            logger.warning(f'Missing {missing_stories} stories')
            logger.info(f'Collected {collected_stories} stories')

    @classmethod
    def splits(cls, root='.data', train='training', validation='validation', test='test', **kwargs):
        path = cls.download(root)
        cls.cache_splits(path)

        train_data = None if train is None else cls(os.path.join(path, 'training.jsonl'), **kwargs)
        validation_data = (
            None if validation is None else cls(os.path.join(path, 'validation.jsonl'), one_answer=False, **kwargs)
        )
        test_data = None if test is None else cls(os.path.join(path, 'test.jsonl'), one_answer=False, **kwargs)

        aux_data = None
        do_curriculum = kwargs.get('curriculum', False)
        if do_curriculum:
            kwargs.pop('curriculum')
            aux_data = cls(os.path.join(path, 'auxiliary.jsonl'), **kwargs)

        return Split(
            train=None if train is None else train_data,
            eval=None if validation is None else validation_data,
            test=None if test is None else test_data,
            aux=None if do_curriculum is None else aux_data,
        )


class DailyMail(Summarization):
    name = 'dailymail'
    dirname = 'dailymail'
    urls = [
        ('https://drive.google.com/uc?export=download&id=0BwmD_VLjROrfM1BxdkxVaTY2bWs', 'dailymail_stories.tgz'),
        (
            'https://raw.githubusercontent.com/abisee/cnn-dailymail/master/url_lists/dailymail_wayback_training_urls.txt',
            'dailymail/dailymail_wayback_training_urls.txt',
        ),
        (
            'https://raw.githubusercontent.com/abisee/cnn-dailymail/master/url_lists/dailymail_wayback_validation_urls.txt',
            'dailymail/dailymail_wayback_validation_urls.txt',
        ),
        (
            'https://raw.githubusercontent.com/abisee/cnn-dailymail/master/url_lists/dailymail_wayback_test_urls.txt',
            'dailymail/dailymail_wayback_test_urls.txt',
        ),
    ]


class CNN(Summarization):
    name = 'cnn'
    dirname = 'cnn'
    urls = [
        ('https://drive.google.com/uc?export=download&id=0BwmD_VLjROrfTHk4NFg2SndKcjQ', 'cnn_stories.tgz'),
        (
            'https://raw.githubusercontent.com/abisee/cnn-dailymail/master/url_lists/cnn_wayback_training_urls.txt',
            'cnn/cnn_wayback_training_urls.txt',
        ),
        (
            'https://raw.githubusercontent.com/abisee/cnn-dailymail/master/url_lists/cnn_wayback_validation_urls.txt',
            'cnn/cnn_wayback_validation_urls.txt',
        ),
        (
            'https://raw.githubusercontent.com/abisee/cnn-dailymail/master/url_lists/cnn_wayback_test_urls.txt',
            'cnn/cnn_wayback_test_urls.txt',
        ),
    ]


class Query:
    # https://github.com/salesforce/WikiSQL/blob/c2ed4f9b22db1cc2721805d53e6e76e07e2ccbdc/lib/query.py#L10

    agg_ops = ['', 'MAX', 'MIN', 'COUNT', 'SUM', 'AVG']
    cond_ops = ['=', '>', '<', 'OP']
    syms = [
        'SELECT',
        'WHERE',
        'AND',
        'COL',
        'TABLE',
        'CAPTION',
        'PAGE',
        'SECTION',
        'OP',
        'COND',
        'QUESTION',
        'AGG',
        'AGGOPS',
        'CONDOPS',
    ]

    def __init__(self, sel_index, agg_index, columns, conditions=tuple()):
        self.sel_index = sel_index
        self.agg_index = agg_index
        self.columns = columns
        self.conditions = list(conditions)

    def __repr__(self):
        rep = 'SELECT {agg} {sel} FROM table'.format(
            agg=self.agg_ops[self.agg_index],
            sel=self.columns[self.sel_index] if self.columns is not None else 'col{}'.format(self.sel_index),
        )
        if self.conditions:
            rep += ' WHERE ' + ' AND '.join(
                ['{} {} {}'.format(self.columns[i], self.cond_ops[o], v) for i, o, v in self.conditions]
            )
        return ' '.join(rep.split())

    @classmethod
    def from_dict(cls, d, t):
        return cls(sel_index=d['sel'], agg_index=d['agg'], columns=t, conditions=d['conds'])


class WikiSQL(CQA):
    urls = ['https://github.com/salesforce/WikiSQL/raw/master/data.tar.bz2']
    name = 'wikisql'
    dirname = 'data'

    def __init__(
        self, path, query_as_question=False, subsample=None, lower=False, cached_path=None, skip_cache=False, **kwargs
    ):
        cache_name = os.path.join(
            cached_path,
            'query_as_question' if query_as_question else 'query_as_context',
            os.path.basename(path),
            str(subsample),
        )
        if os.path.exists(cache_name) and not skip_cache:
            logger.info(f'Loading cached data from {cache_name}')
            examples, all_answers = torch.load(cache_name)
        else:

            expanded_path = os.path.expanduser(path)
            table_path = os.path.splitext(expanded_path)
            table_path = table_path[0] + '.tables' + table_path[1]

            with open(table_path) as tables_file:
                tables = [json.loads(line) for line in tables_file]
                id_to_tables = {x['id']: x for x in tables}

            all_answers = []
            examples = []
            with open(expanded_path) as example_file:
                for idx, line in enumerate(example_file):
                    entry = json.loads(line)
                    human_query = entry['question']
                    table = id_to_tables[entry['table_id']]
                    sql = entry['sql']
                    header = table['header']
                    answer = repr(Query.from_dict(sql, header))
                    context = (
                        f'The table has columns {", ".join(table["header"])} '
                        + f'and key words {", ".join(Query.agg_ops[1:] + Query.cond_ops + Query.syms)}'
                    )
                    if query_as_question:
                        question = human_query
                    else:
                        question = 'What is the translation from English to SQL?'
                        context += f'-- {human_query}'
                    examples.append(Example.from_raw(make_example_id(self, idx), context, question, answer, lower=lower))
                    all_answers.append({'sql': sql, 'header': header, 'answer': answer, 'table': table})
                    if subsample is not None and len(examples) > subsample:
                        break

            os.makedirs(os.path.dirname(cache_name), exist_ok=True)
            logger.info(f'Caching data to {cache_name}')
            torch.save((examples, all_answers), cache_name)

        super(WikiSQL, self).__init__(examples, **kwargs)
        self.all_answers = all_answers

    @classmethod
    def splits(cls, root='.data', train='train.jsonl', validation='dev.jsonl', test='test.jsonl', **kwargs):
        """Create dataset objects for splits of the SQuAD dataset.
        Arguments:
            root: directory containing SQuAD data
            field: field for handling all columns
            train: The prefix of the train data. Default: 'train'.
            validation: The prefix of the validation data. Default: 'val'.
            Remaining keyword arguments: Passed to the splits method of
                Dataset.
        """
        path = cls.download(root)

        train_data = None if train is None else cls(os.path.join(path, train), **kwargs)
        validation_data = None if validation is None else cls(os.path.join(path, validation), **kwargs)
        test_data = None if test is None else cls(os.path.join(path, test), **kwargs)

        aux_data = None
        do_curriculum = kwargs.get('curriculum', False)
        if do_curriculum:
            kwargs.pop('curriculum')
            aux_data = cls(os.path.join(path, 'aux'), **kwargs)

        return Split(
            train=None if train is None else train_data,
            eval=None if validation is None else validation_data,
            test=None if test is None else test_data,
            aux=None if do_curriculum is None else aux_data,
        )


class SRL(CQA):
    urls = [
        'https://dada.cs.washington.edu/qasrl/data/wiki1.train.qa',
        'https://dada.cs.washington.edu/qasrl/data/wiki1.dev.qa',
        'https://dada.cs.washington.edu/qasrl/data/wiki1.test.qa',
    ]

    name = 'srl'
    dirname = ''

    @classmethod
    def clean(cls, s):
        closing_punctuation = {
            ' .',
            ' ,',
            ' ;',
            ' !',
            ' ?',
            ' :',
            ' )',
            " 'll",
            " n't ",
            " %",
            " 't",
            " 's",
            " 'm",
            " 'd",
            " 're",
        }
        opening_punctuation = {'( ', '$ '}
        both_sides = {' - '}
        s = ' '.join(s.split()).strip()
        s = s.replace('-LRB-', '(')
        s = s.replace('-RRB-', ')')
        s = s.replace('-LAB-', '<')
        s = s.replace('-RAB-', '>')
        s = s.replace('-AMP-', '&')
        s = s.replace('%pw', ' ')

        for p in closing_punctuation:
            s = s.replace(p, p.lstrip())
        for p in opening_punctuation:
            s = s.replace(p, p.rstrip())
        for p in both_sides:
            s = s.replace(p, p.strip())
        s = s.replace('``', '')
        s = s.replace('`', '')
        s = s.replace("''", '')
        s = s.replace('“', '')
        s = s.replace('”', '')
        s = s.replace(" '", '')
        return ' '.join(s.split()).strip()

    def __init__(self, path, one_answer=True, subsample=None, lower=False, cached_path=None, skip_cache=False, **kwargs):
        cache_name = os.path.join(cached_path, os.path.basename(path), str(subsample))

        examples, all_answers = [], []
        if os.path.exists(cache_name) and not skip_cache:
            logger.info(f'Loading cached data from {cache_name}')
            examples, all_answers = torch.load(cache_name)
        else:
            with open(os.path.expanduser(path)) as f:
                for line in f:
                    ex = json.loads(line)
                    aa = ex['all_answers']
                    context, question, answer = ex['context'], ex['question'], ex['answer']
                    examples.append(
                        Example.from_raw(make_example_id(self, len(all_answers)), context, question, answer, lower=lower)
                    )
                    all_answers.append(aa)
                    if subsample is not None and len(examples) >= subsample:
                        break
            os.makedirs(os.path.dirname(cache_name), exist_ok=True)
            logger.info(f'Caching data to {cache_name}')
            torch.save((examples, all_answers), cache_name)

        super(SRL, self).__init__(examples, **kwargs)
        self.all_answers = all_answers

    @classmethod
    def cache_splits(cls, path, train='train', validation='dev', test='test'):

        splits = [train, validation, test]
        for split in splits:
            split_file_name = os.path.join(path, f'{split}.jsonl')
            if os.path.exists(split_file_name):
                continue
            wiki_file = os.path.join(path, f'wiki1.{split}.qa')

            with open(split_file_name, 'w') as split_file:
                with open(os.path.expanduser(wiki_file)) as f:

                    def is_int(x):
                        try:
                            int(x)
                            return True
                        except ValueError:
                            return False

                    lines = []
                    for line in f.readlines():
                        line = ' '.join(line.split()).strip()
                        if len(line) == 0:
                            lines.append(line)
                            continue
                        if 'WIKI1' not in line.split('_')[0]:
                            if not is_int(line.split()[0]) or len(line.split()) > 3:
                                lines.append(line)

                    new_example = True
                    for line in lines:
                        line = line.strip()
                        if new_example:
                            context = cls.clean(line)
                            new_example = False
                            continue
                        if len(line) == 0:
                            new_example = True
                            continue
                        question, answers = line.split('?')
                        question = cls.clean(line.split('?')[0].replace(' _', '') + '?')
                        answer = cls.clean(answers.split('###')[0])
                        all_answers = [cls.clean(x) for x in answers.split('###')]
                        if answer not in context:
                            low_answer = answer[0].lower() + answer[1:]
                            up_answer = answer[0].upper() + answer[1:]
                            if low_answer in context or up_answer in context:
                                answer = low_answer if low_answer in context else up_answer
                            else:
                                if 'Darcy Burner' in answer:
                                    answer = 'Darcy Burner and other 2008 Democratic congressional candidates, in cooperation with some retired national security officials'
                                elif 'E Street Band' in answer:
                                    answer = 'plan to work with the E Street Band again in the future'
                                elif 'an electric sender' in answer:
                                    answer = 'an electronic sender'
                                elif 'the US army' in answer:
                                    answer = 'the US Army'
                                elif 'Rather than name the' in answer:
                                    answer = 'rather die than name the cause of his disease to his father'
                                elif answer.lower() in context:
                                    answer = answer.lower()
                                else:
                                    import pdb

                                    pdb.set_trace()
                        assert answer in context
                        modified_all_answers = []
                        for a in all_answers:
                            if a not in context:
                                low_answer = a[0].lower() + a[1:]
                                up_answer = a[0].upper() + a[1:]
                                if low_answer in context or up_answer in context:
                                    a = low_answer if low_answer in context else up_answer
                                else:
                                    if 'Darcy Burner' in a:
                                        a = 'Darcy Burner and other 2008 Democratic congressional candidates, in cooperation with some retired national security officials'
                                    elif 'E Street Band' in a:
                                        a = 'plan to work with the E Street Band again in the future'
                                    elif 'an electric sender' in a:
                                        a = 'an electronic sender'
                                    elif 'the US army' in a:
                                        a = 'the US Army'
                                    elif 'Rather than name the' in a:
                                        a = 'rather die than name the cause of his disease to his father'
                                    elif a.lower() in context:
                                        a = a.lower()
                                    else:
                                        import pdb

                                        pdb.set_trace()
                            assert a in context
                            modified_all_answers.append(a)
                        split_file.write(
                            json.dumps(
                                {
                                    'context': context,
                                    'question': question,
                                    'answer': answer,
                                    'type': 'wiki',
                                    'all_answers': modified_all_answers,
                                }
                            )
                            + '\n'
                        )

    @classmethod
    def splits(cls, root='.data', train='train', validation='dev', test='test', **kwargs):
        path = cls.download(root)
        cls.cache_splits(path)

        train_data = None if train is None else cls(os.path.join(path, f'{train}.jsonl'), **kwargs)
        validation_data = (
            None if validation is None else cls(os.path.join(path, f'{validation}.jsonl'), one_answer=False, **kwargs)
        )
        test_data = None if test is None else cls(os.path.join(path, f'{test}.jsonl'), one_answer=False, **kwargs)
        aux_data = None
        do_curriculum = kwargs.get('curriculum', False)
        if do_curriculum:
            kwargs.pop('curriculum')
            aux_data = cls(os.path.join(path, 'aux.jsonl'), **kwargs)

        return Split(
            train=None if train is None else train_data,
            eval=None if validation is None else validation_data,
            test=None if test is None else test_data,
            aux=None if do_curriculum is None else aux_data,
        )


class WinogradSchema(CQA):
    urls = ['https://s3.amazonaws.com/research.metamind.io/decaNLP/data/schema.txt']

    name = 'schema'
    dirname = ''

    def __init__(self, path, subsample=None, lower=False, cached_path=None, skip_cache=False, **kwargs):
        cache_name = os.path.join(cached_path, os.path.basename(path), str(subsample))
        if os.path.exists(cache_name) and not skip_cache:
            logger.info(f'Loading cached data from {cache_name}')
            examples = torch.load(cache_name)
        else:
            examples = []
            with open(os.path.expanduser(path)) as f:
                for line in f:
                    ex = json.loads(line)
                    context, question, answer = ex['context'], ex['question'], ex['answer']
                    examples.append(
                        Example.from_raw(make_example_id(self, len(examples)), context, question, answer, lower=lower)
                    )
                    if subsample is not None and len(examples) >= subsample:
                        break
            os.makedirs(os.path.dirname(cache_name), exist_ok=True)
            logger.info(f'Caching data to {cache_name}')
            torch.save(examples, cache_name)

        super(WinogradSchema, self).__init__(examples, **kwargs)

    @classmethod
    def cache_splits(cls, path):
        pattern = '\[.*\]'
        train_jsonl = os.path.expanduser(os.path.join(path, 'train.jsonl'))
        if os.path.exists(train_jsonl):
            return

        def get_both_schema(context):
            variations = [x[1:-1].split('/') for x in re.findall(pattern, context)]
            splits = re.split(pattern, context)
            results = []
            for which_schema in range(2):
                vs = [v[which_schema] for v in variations]
                context = ''
                for idx in range(len(splits)):
                    context += splits[idx]
                    if idx < len(vs):
                        context += vs[idx]
                results.append(context)
            return results

        schemas = []
        with open(os.path.expanduser(os.path.join(path, 'schema.txt'))) as schema_file:
            schema = []
            for line in schema_file:
                if len(line.split()) == 0:
                    schemas.append(schema)
                    schema = []
                    continue
                else:
                    schema.append(line.strip())

        examples = []
        for schema in schemas:
            context, question, answer = schema
            contexts = get_both_schema(context)
            questions = get_both_schema(question)
            answers = answer.split('/')
            for idx in range(2):
                answer = answers[idx]
                question = questions[idx] + f' {answers[0]} or {answers[1]}?'
                examples.append({'context': contexts[idx], 'question': question, 'answer': answer})

        traindev = examples[:-100]
        test = examples[-100:]
        train = traindev[:80]
        dev = traindev[80:]

        splits = ['train', 'validation', 'test']
        for split, examples in zip(splits, [train, dev, test]):
            with open(os.path.expanduser(os.path.join(path, f'{split}.jsonl')), 'a') as split_file:
                for ex in examples:
                    split_file.write(json.dumps(ex) + '\n')

    @classmethod
    def splits(cls, root='.data', train='train', validation='validation', test='test', **kwargs):
        path = cls.download(root)
        cls.cache_splits(path)

        train_data = None if train is None else cls(os.path.join(path, f'{train}.jsonl'), **kwargs)
        validation_data = None if validation is None else cls(os.path.join(path, f'{validation}.jsonl'), **kwargs)
        test_data = None if test is None else cls(os.path.join(path, f'{test}.jsonl'), **kwargs)

        aux_data = None
        do_curriculum = kwargs.get('curriculum', False)
        if do_curriculum:
            kwargs.pop('curriculum')
            aux_data = cls(os.path.join(path, 'aux.jsonl'), **kwargs)

        return Split(
            train=None if train is None else train_data,
            eval=None if validation is None else validation_data,
            test=None if test is None else test_data,
            aux=None if do_curriculum is None else aux_data,
        )


class WOZ(CQA):
    urls = [
        'https://raw.githubusercontent.com/nmrksic/neural-belief-tracker/master/data/woz/woz_train_en.json',
        'https://raw.githubusercontent.com/nmrksic/neural-belief-tracker/master/data/woz/woz_test_de.json',
        'https://raw.githubusercontent.com/nmrksic/neural-belief-tracker/master/data/woz/woz_test_en.json',
        'https://raw.githubusercontent.com/nmrksic/neural-belief-tracker/master/data/woz/woz_train_de.json',
        'https://raw.githubusercontent.com/nmrksic/neural-belief-tracker/master/data/woz/woz_train_en.json',
        'https://raw.githubusercontent.com/nmrksic/neural-belief-tracker/master/data/woz/woz_validate_de.json',
        'https://raw.githubusercontent.com/nmrksic/neural-belief-tracker/master/data/woz/woz_validate_en.json',
    ]

    name = 'woz'
    dirname = ''

    def __init__(self, path, subsample=None, lower=False, description='woz.en', cached_path=None, skip_cache=False, **kwargs):
        examples, all_answers = [], []
        cache_name = os.path.join(cached_path, os.path.basename(path), str(subsample), description)
        if os.path.exists(cache_name) and not skip_cache:
            logger.info(f'Loading cached data from {cache_name}')
            examples, all_answers = torch.load(cache_name)
        else:
            with open(os.path.expanduser(path)) as f:
                for woz_id, line in enumerate(f):
                    ex = example_dict = json.loads(line)
                    if example_dict['lang'] in description:
                        context, question, answer = ex['context'], ex['question'], ex['answer']
                        all_answers.append((ex['lang_dialogue_turn'], answer))
                        examples.append(
                            Example.from_raw(make_example_id(self, woz_id), context, question, answer, lower=lower)
                        )

                    if subsample is not None and len(examples) >= subsample:
                        break
            os.makedirs(os.path.dirname(cache_name), exist_ok=True)
            logger.info(f'Caching data to {cache_name}')
            torch.save((examples, all_answers), cache_name)

        super(WOZ, self).__init__(examples, **kwargs)
        self.all_answers = all_answers

    @classmethod
    def cache_splits(cls, path, train='train', validation='validate', test='test'):
        train_jsonl = os.path.expanduser(os.path.join(path, 'train.jsonl'))
        if os.path.exists(train_jsonl):
            return

        file_name_base = 'woz_{}_{}.json'
        for split in [train, validation, test]:
            with open(os.path.expanduser(os.path.join(path, f'{split}.jsonl')), 'a') as split_file:
                for lang in ['en', 'de']:
                    file_path = file_name_base.format(split, lang)
                    with open(os.path.expanduser(os.path.join(path, file_path))) as src_file:
                        dialogues = json.loads(src_file.read())
                        for di, d in enumerate(dialogues):
                            previous_state = {'inform': [], 'request': []}
                            turns = d['dialogue']
                            for ti, t in enumerate(turns):
                                question = 'What is the change in state?'
                                actions = []
                                for act in t['system_acts']:
                                    if isinstance(act, list):
                                        act = ': '.join(act)
                                    actions.append(act)
                                actions = ', '.join(actions)
                                if len(actions) > 0:
                                    actions += ' -- '
                                context = actions + t['transcript']
                                belief_state = t['belief_state']
                                delta_state = {'inform': [], 'request': []}
                                current_state = {'inform': [], 'request': []}
                                for item in belief_state:
                                    if 'slots' in item:
                                        slots = item['slots']
                                        for slot in slots:
                                            act = item['act']
                                            if act == 'inform':
                                                current_state['inform'].append(slot)
                                                if slot not in previous_state['inform']:
                                                    delta_state['inform'].append(slot)
                                                else:
                                                    prev_slot = previous_state['inform'][previous_state['inform'].index(slot)]
                                                    if prev_slot[1] != slot[1]:
                                                        delta_state['inform'].append(slot)
                                            else:
                                                delta_state['request'].append(slot[1])
                                                current_state['request'].append(slot[1])
                                previous_state = current_state
                                answer = ''
                                if len(delta_state['inform']) > 0:
                                    answer = ', '.join([f'{x[0]}: {x[1]}' for x in delta_state['inform']])
                                answer += ';'
                                if len(delta_state['request']) > 0:
                                    answer += ' '
                                    answer += ', '.join(delta_state['request'])
                                ex = {
                                    'context': ' '.join(context.split()),
                                    'question': ' '.join(question.split()),
                                    'lang': lang,
                                    'answer': answer if len(answer) > 1 else 'None',
                                    'lang_dialogue_turn': f'{lang}_{di}_{ti}',
                                }
                                split_file.write(json.dumps(ex) + '\n')

    @classmethod
    def splits(cls, root='.data', train='train', validation='validate', test='test', **kwargs):
        path = cls.download(root)
        cls.cache_splits(path)

        train_data = None if train is None else cls(os.path.join(path, f'{train}.jsonl'), **kwargs)
        validation_data = None if validation is None else cls(os.path.join(path, f'{validation}.jsonl'), **kwargs)
        test_data = None if test is None else cls(os.path.join(path, f'{test}.jsonl'), **kwargs)

        aux_data = None
        do_curriculum = kwargs.get('curriculum', False)
        if do_curriculum:
            kwargs.pop('curriculum')
            aux_data = cls(os.path.join(path, 'aux.jsonl'), **kwargs)

        return Split(
            train=None if train is None else train_data,
            eval=None if validation is None else validation_data,
            test=None if test is None else test_data,
            aux=None if do_curriculum is None else aux_data,
        )


class MultiNLI(CQA):
    urls = ['http://www.nyu.edu/projects/bowman/multinli/multinli_1.0.zip']

    name = 'multinli'
    dirname = 'multinli_1.0'

    def __init__(
        self, path, subsample=None, lower=False, description='multinli.in.out', cached_path=None, skip_cache=False, **kwargs
    ):
        cache_name = os.path.join(cached_path, os.path.basename(path), str(subsample), description)
        if os.path.exists(cache_name) and not skip_cache:
            logger.info(f'Loading cached data from {cache_name}')
            examples = torch.load(cache_name)
        else:
            examples = []
            with open(os.path.expanduser(path)) as f:
                for line in f:
                    ex = example_dict = json.loads(line)
                    if example_dict['subtask'] in description:
                        context, question, answer = ex['context'], ex['question'], ex['answer']
                        examples.append(
                            Example.from_raw(make_example_id(self, len(examples)), context, question, answer, lower=lower)
                        )
                    if subsample is not None and len(examples) >= subsample:
                        break
            os.makedirs(os.path.dirname(cache_name), exist_ok=True)
            logger.info(f'Caching data to {cache_name}')
            torch.save(examples, cache_name)

        super(MultiNLI, self).__init__(examples, **kwargs)

    @classmethod
    def cache_splits(cls, path, train='multinli_1.0_train', validation='mulinli_1.0_dev_{}', test='test'):
        train_jsonl = os.path.expanduser(os.path.join(path, 'train.jsonl'))
        if os.path.exists(train_jsonl):
            return

        with open(os.path.expanduser(os.path.join(path, 'train.jsonl')), 'a') as split_file:
            with open(os.path.expanduser(os.path.join(path, 'multinli_1.0_train.jsonl'))) as src_file:
                for line in src_file:
                    ex = json.loads(line)
                    ex = {
                        'context': f'Premise: "{ex["sentence1"]}"',
                        'question': f'Hypothesis: "{ex["sentence2"]}" -- entailment, neutral, or contradiction?',
                        'answer': ex['gold_label'],
                        'subtask': 'multinli',
                    }
                    split_file.write(json.dumps(ex) + '\n')

        with open(os.path.expanduser(os.path.join(path, 'validation.jsonl')), 'a') as split_file:
            for subtask in ['matched', 'mismatched']:
                with open(os.path.expanduser(os.path.join(path, 'multinli_1.0_dev_{}.jsonl'.format(subtask)))) as src_file:
                    for line in src_file:
                        ex = json.loads(line)
                        ex = {
                            'context': f'Premise: "{ex["sentence1"]}"',
                            'question': f'Hypothesis: "{ex["sentence2"]}" -- entailment, neutral, or contradiction?',
                            'answer': ex['gold_label'],
                            'subtask': 'in' if subtask == 'matched' else 'out',
                        }
                        split_file.write(json.dumps(ex) + '\n')

    @classmethod
    def splits(cls, root='.data', train='train', validation='validation', test='test', **kwargs):
        path = cls.download(root)
        cls.cache_splits(path)

        train_data = None if train is None else cls(os.path.join(path, f'{train}.jsonl'), **kwargs)
        validation_data = None if validation is None else cls(os.path.join(path, f'{validation}.jsonl'), **kwargs)
        test_data = None if test is None else cls(os.path.join(path, f'{test}.jsonl'), **kwargs)

        aux_data = None
        do_curriculum = kwargs.get('curriculum', False)
        if do_curriculum:
            kwargs.pop('curriculum')
            aux_data = cls(os.path.join(path, 'aux.jsonl'), **kwargs)

        return Split(
            train=None if train is None else train_data,
            eval=None if validation is None else validation_data,
            test=None if test is None else test_data,
            aux=None if do_curriculum is None else aux_data,
        )


class ZeroShotRE(CQA):
    urls = ['http://nlp.cs.washington.edu/zeroshot/relation_splits.tar.bz2']
    dirname = 'relation_splits'
    name = 'zre'

    def __init__(self, path, subsample=None, lower=False, cached_path=None, skip_cache=False, **kwargs):
        cache_name = os.path.join(cached_path, os.path.basename(path), str(subsample))
        if os.path.exists(cache_name) and not skip_cache:
            logger.info(f'Loading cached data from {cache_name}')
            examples = torch.load(cache_name)
        else:
            examples = []
            with open(os.path.expanduser(path)) as f:
                for line in f:
                    ex = json.loads(line)
                    context, question, answer = ex['context'], ex['question'], ex['answer']
                    examples.append(
                        Example.from_raw(make_example_id(self, len(examples)), context, question, answer, lower=lower)
                    )

                    if subsample is not None and len(examples) >= subsample:
                        break
            os.makedirs(os.path.dirname(cache_name), exist_ok=True)
            logger.info(f'Caching data to {cache_name}')
            torch.save(examples, cache_name)

        super().__init__(examples, **kwargs)

    @classmethod
    def cache_splits(cls, path, train='train', validation='dev', test='test'):
        train_jsonl = os.path.expanduser(os.path.join(path, f'{train}.jsonl'))
        if os.path.exists(train_jsonl):
            return

        base_file_name = '{}.0'
        for split in [train, validation, test]:
            src_file_name = base_file_name.format(split)
            with open(os.path.expanduser(os.path.join(path, f'{split}.jsonl')), 'a') as split_file:
                with open(os.path.expanduser(os.path.join(path, src_file_name))) as src_file:
                    for line in src_file:
                        split_line = line.split('\t')
                        if len(split_line) == 4:
                            answer = ''
                            relation, question, subject, context = split_line
                        else:
                            relation, question, subject, context = split_line[:4]
                            answer = ', '.join(split_line[4:])
                        question = question.replace('XXX', subject)
                        ex = {
                            'context': context,
                            'question': question,
                            'answer': answer if len(answer) > 0 else 'unanswerable',
                        }
                        split_file.write(json.dumps(ex) + '\n')

    @classmethod
    def splits(cls, root='.data', train='train', validation='dev', test='test', **kwargs):
        path = cls.download(root)
        cls.cache_splits(path)

        train_data = None if train is None else cls(os.path.join(path, f'{train}.jsonl'), **kwargs)
        validation_data = None if validation is None else cls(os.path.join(path, f'{validation}.jsonl'), **kwargs)
        test_data = None if test is None else cls(os.path.join(path, f'{test}.jsonl'), **kwargs)

        aux_data = None
        do_curriculum = kwargs.get('curriculum', False)
        if do_curriculum:
            kwargs.pop('curriculum')
            aux_data = cls(os.path.join(path, 'aux.jsonl'), **kwargs)

        return Split(
            train=None if train is None else train_data,
            eval=None if validation is None else validation_data,
            test=None if test is None else test_data,
            aux=None if do_curriculum is None else aux_data,
        )


class OntoNotesNER(CQA):
    urls = [
        'http://conll.cemantix.org/2012/download/ids/english/all/train.id',
        'http://conll.cemantix.org/2012/download/ids/english/all/development.id',
        'http://conll.cemantix.org/2012/download/ids/english/all/test.id',
    ]

    name = 'ontonotes.ner'
    dirname = ''

    @classmethod
    def clean(cls, s):
        closing_punctuation = {' .', ' ,', ' ;', ' !', ' ?', ' :', ' )', " '", " n't ", " %"}
        opening_punctuation = {'( ', '$ '}
        both_sides = {' - '}
        s = ' '.join(s.split()).strip()
        s = s.replace(' /.', ' .')
        s = s.replace(' /?', ' ?')
        s = s.replace('-LRB-', '(')
        s = s.replace('-RRB-', ')')
        s = s.replace('-LAB-', '<')
        s = s.replace('-RAB-', '>')
        s = s.replace('-AMP-', '&')
        s = s.replace('%pw', ' ')

        for p in closing_punctuation:
            s = s.replace(p, p.lstrip())
        for p in opening_punctuation:
            s = s.replace(p, p.rstrip())
        for p in both_sides:
            s = s.replace(p, p.strip())
        s = s.replace('``', '"')
        s = s.replace("''", '"')
        quote_is_open = True
        quote_idx = s.find('"')
        raw = ''
        while quote_idx >= 0:
            start_enamex_open_idx = s.find('<ENAMEX')
            if start_enamex_open_idx > -1:
                end_enamex_open_idx = s.find('">') + 2
                if start_enamex_open_idx <= quote_idx <= end_enamex_open_idx:
                    raw += s[:end_enamex_open_idx]
                    s = s[end_enamex_open_idx:]
                    quote_idx = s.find('"')
                    continue
            if quote_is_open:
                raw += s[: quote_idx + 1]
                s = s[quote_idx + 1 :].strip()
                quote_is_open = False
            else:
                raw += s[:quote_idx].strip() + '"'
                s = s[quote_idx + 1 :]
                quote_is_open = True
            quote_idx = s.find('"')
        raw += s

        return ' '.join(raw.split()).strip()

    def __init__(
        self,
        path,
        one_answer=True,
        subsample=None,
        lower=False,
        path_to_files='.data/ontonotes-release-5.0/data/files',
        subtask='all',
        nones=True,
        cached_path=None,
        skip_cache=False,
        **kwargs,
    ):
        cache_name = os.path.join(cached_path, os.path.basename(path), str(subsample), subtask, str(nones))
        if os.path.exists(cache_name) and not skip_cache:
            logger.info(f'Loading cached data from {cache_name}')
            examples = torch.load(cache_name)
        else:
            examples = []
            with open(os.path.expanduser(path)) as f:
                for line in f:
                    example_dict = json.loads(line)
                    t = example_dict['type']
                    a = example_dict['answer']
                    if subtask == 'both' or t == subtask:
                        if a != 'None' or nones:
                            ex = example_dict
                            context, question, answer = ex['context'], ex['question'], ex['answer']
                            examples.append(
                                Example.from_raw(make_example_id(self, len(examples)), context, question, answer, lower=lower)
                            )

                    if subsample is not None and len(examples) >= subsample:
                        break
            os.makedirs(os.path.dirname(cache_name), exist_ok=True)
            logger.info(f'Caching data to {cache_name}')
            torch.save(examples, cache_name)

        super(OntoNotesNER, self).__init__(examples, **kwargs)

    @classmethod
    def cache_splits(cls, path, path_to_files, train='train', validation='development', test='test'):

        label_to_answer = {
            'PERSON': 'person',
            'NORP': 'political',
            'FAC': 'facility',
            'ORG': 'organization',
            'GPE': 'geopolitical',
            'LOC': 'location',
            'PRODUCT': 'product',
            'EVENT': 'event',
            'WORK_OF_ART': 'artwork',
            'LAW': 'legal',
            'LANGUAGE': 'language',
            'DATE': 'date',
            'TIME': 'time',
            'PERCENT': 'percentage',
            'MONEY': 'monetary',
            'QUANTITY': 'quantitative',
            'ORDINAL': 'ordinal',
            'CARDINAL': 'cardinal',
        }

        pluralize = {
            'person': 'persons',
            'political': 'political',
            'facility': 'facilities',
            'organization': 'organizations',
            'geopolitical': 'geopolitical',
            'location': 'locations',
            'product': 'products',
            'event': 'events',
            'artwork': 'artworks',
            'legal': 'legal',
            'language': 'languages',
            'date': 'dates',
            'time': 'times',
            'percentage': 'percentages',
            'monetary': 'monetary',
            'quantitative': 'quantitative',
            'ordinal': 'ordinal',
            'cardinal': 'cardinal',
        }

        for split in [train, validation, test]:
            split_file_name = os.path.join(path, f'{split}.jsonl')
            if os.path.exists(split_file_name):
                continue
            id_file = os.path.join(path, f'{split}.id')

            num_file_ids = 0
            with open(split_file_name, 'w') as split_file:
                with open(os.path.expanduser(id_file)) as f:
                    for file_id in f:
                        example_file_name = os.path.join(os.path.expanduser(path_to_files), file_id.strip()) + '.name'
                        if not os.path.exists(example_file_name) or 'annotations/tc/ch' in example_file_name:
                            continue
                        num_file_ids += 1
                        with open(example_file_name) as example_file:
                            lines = [x.strip() for x in example_file.readlines() if 'DOC' not in x]
                            for line in lines:
                                original = line
                                line = cls.clean(line)
                                entities = []
                                while True:
                                    start_enamex_open_idx = line.find('<ENAMEX')
                                    if start_enamex_open_idx == -1:
                                        break
                                    end_enamex_open_idx = line.find('">') + 2
                                    start_enamex_close_idx = line.find('</ENAMEX>')
                                    end_enamex_close_idx = start_enamex_close_idx + len('</ENAMEX>')

                                    enamex_open_tag = line[start_enamex_open_idx:end_enamex_open_idx]
                                    before_entity = line[:start_enamex_open_idx]
                                    entity = line[end_enamex_open_idx:start_enamex_close_idx]
                                    after_entity = line[end_enamex_close_idx:]

                                    if 'S_OFF' in enamex_open_tag:
                                        s_off_start = enamex_open_tag.find('S_OFF="')
                                        s_off_end = (
                                            enamex_open_tag.find('">')
                                            if 'E_OFF' not in enamex_open_tag
                                            else enamex_open_tag.find('" E_OFF')
                                        )
                                        s_off = int(enamex_open_tag[s_off_start + len('S_OFF="') : s_off_end])
                                        enamex_open_tag = enamex_open_tag[: s_off_start - 2] + '">'
                                        before_entity += entity[:s_off]
                                        entity = entity[s_off:]

                                    if 'E_OFF' in enamex_open_tag:
                                        s_off_start = enamex_open_tag.find('E_OFF="')
                                        s_off_end = enamex_open_tag.find('">')
                                        s_off = int(enamex_open_tag[s_off_start + len('E_OFF="') : s_off_end])
                                        enamex_open_tag = enamex_open_tag[: s_off_start - 2] + '">'
                                        after_entity = entity[-s_off:] + after_entity
                                        entity = entity[:-s_off]

                                    label_start = enamex_open_tag.find('TYPE="') + len('TYPE="')
                                    label_end = enamex_open_tag.find('">')
                                    label = enamex_open_tag[label_start:label_end]
                                    assert label in label_to_answer

                                    offsets = (len(before_entity), len(before_entity) + len(entity))
                                    entities.append({'entity': entity, 'char_offsets': offsets, 'label': label})
                                    line = before_entity + entity + after_entity

                                context = line.strip()
                                is_no_good = False
                                for entity_tuple in entities:
                                    entity = entity_tuple['entity']
                                    start, end = entity_tuple['char_offsets']
                                    if not context[start:end] == entity:
                                        is_no_good = True
                                        break
                                if is_no_good:
                                    logger.warning(
                                        'Throwing out example that looks poorly labeled: ',
                                        original.strip(),
                                        ' (',
                                        file_id.strip(),
                                        ')',
                                    )
                                    continue
                                question = 'What are the tags for all entities?'
                                answer = '; '.join([f'{x["entity"]} -- {label_to_answer[x["label"]]}' for x in entities])
                                if len(answer) == 0:
                                    answer = 'None'
                                split_file.write(
                                    json.dumps(
                                        {
                                            'context': context,
                                            'question': question,
                                            'answer': answer,
                                            'file_id': file_id.strip(),
                                            'original': original.strip(),
                                            'entity_list': entities,
                                            'type': 'all',
                                        }
                                    )
                                    + '\n'
                                )
                                partial_question = 'Which entities are {}?'

                                for lab, ans in label_to_answer.items():
                                    question = partial_question.format(pluralize[ans])
                                    entity_of_type_lab = [x['entity'] for x in entities if x['label'] == lab]
                                    answer = ', '.join(entity_of_type_lab)
                                    if len(answer) == 0:
                                        answer = 'None'
                                    split_file.write(
                                        json.dumps(
                                            {
                                                'context': context,
                                                'question': question,
                                                'answer': answer,
                                                'file_id': file_id.strip(),
                                                'original': original.strip(),
                                                'entity_list': entities,
                                                'type': 'one',
                                            }
                                        )
                                        + '\n'
                                    )

    @classmethod
    def splits(cls, root='.data', train='train', validation='development', test='test', **kwargs):
        path_to_files = os.path.join(root, 'ontonotes-release-5.0', 'data', 'files')
        assert os.path.exists(path_to_files)
        path = cls.download(root)
        cls.cache_splits(path, path_to_files)

        train_data = None if train is None else cls(os.path.join(path, f'{train}.jsonl'), **kwargs)
        validation_data = (
            None if validation is None else cls(os.path.join(path, f'{validation}.jsonl'), one_answer=False, **kwargs)
        )
        test_data = None if test is None else cls(os.path.join(path, f'{test}.jsonl'), one_answer=False, **kwargs)

        aux_data = None
        do_curriculum = kwargs.get('curriculum', False)
        if do_curriculum:
            kwargs.pop('curriculum')
            aux_data = cls(os.path.join(path, 'aux.jsonl'), **kwargs)

        return Split(
            train=None if train is None else train_data,
            eval=None if validation is None else validation_data,
            test=None if test is None else test_data,
            aux=None if do_curriculum is None else aux_data,
        )


class SNLI(CQA):
    urls = ['http://nlp.stanford.edu/projects/snli/snli_1.0.zip']
    dirname = 'snli_1.0'
    name = 'snli'

    def __init__(self, path, subsample=None, lower=False, cached_path=None, skip_cache=False, **kwargs):
        cache_name = os.path.join(cached_path, os.path.basename(path), str(subsample))
        if os.path.exists(cache_name) and not skip_cache:
            logger.info(f'Loading cached data from {cache_name}')
            examples = torch.load(cache_name)
        else:
            examples = []
            with open(os.path.expanduser(path)) as f:
                for line in f:
                    example_dict = json.loads(line)
                    ex = example_dict
                    context, question, answer = ex['context'], ex['question'], ex['answer']
                    examples.append(
                        Example.from_raw(make_example_id(self, len(examples)), context, question, answer, lower=lower)
                    )

                    if subsample is not None and len(examples) >= subsample:
                        break
            os.makedirs(os.path.dirname(cache_name), exist_ok=True)
            logger.info(f'Caching data to {cache_name}')
            torch.save(examples, cache_name)

        super().__init__(examples, **kwargs)

    @classmethod
    def cache_splits(cls, path, train='train', validation='dev', test='test'):
        train_jsonl = os.path.expanduser(os.path.join(path, f'{train}.jsonl'))
        if os.path.exists(train_jsonl):
            return

        base_file_name = 'snli_1.0_{}.jsonl'
        for split in [train, validation, test]:
            src_file_name = base_file_name.format(split)
            with open(os.path.expanduser(os.path.join(path, f'{split}.jsonl')), 'a') as split_file:
                with open(os.path.expanduser(os.path.join(path, src_file_name))) as src_file:
                    for line in src_file:
                        ex = json.loads(line)
                        ex = {
                            'context': f'Premise: "{ex["sentence1"]}"',
                            'question': f'Hypothesis: "{ex["sentence2"]}" -- entailment, neutral, or contradiction?',
                            'answer': ex['gold_label'],
                        }
                        split_file.write(json.dumps(ex) + '\n')

    @classmethod
    def splits(cls, root='.data', train='train', validation='dev', test='test', **kwargs):
        path = cls.download(root)
        cls.cache_splits(path)

        train_data = None if train is None else cls(os.path.join(path, f'{train}.jsonl'), **kwargs)
        validation_data = None if validation is None else cls(os.path.join(path, f'{validation}.jsonl'), **kwargs)
        test_data = None if test is None else cls(os.path.join(path, f'{test}.jsonl'), **kwargs)

        aux_data = None
        do_curriculum = kwargs.get('curriculum', False)
        if do_curriculum:
            kwargs.pop('curriculum')
            aux_data = cls(os.path.join(path, 'aux.jsonl'), **kwargs)

        return Split(
            train=None if train is None else train_data,
            eval=None if validation is None else validation_data,
            test=None if test is None else test_data,
            aux=None if do_curriculum is None else aux_data,
        )


class JSON(CQA):
    name = 'json'

    def __init__(self, path, subsample=None, lower=False, cached_path=None, skip_cache=False, **kwargs):
        cache_name = os.path.join(cached_path, os.path.basename(path), str(subsample))

        examples = []
        if os.path.exists(cache_name) and not skip_cache:
            logger.info(f'Loading cached data from {cache_name}')
            examples = torch.load(cache_name)
        else:
            with open(os.path.expanduser(path)) as f:
                lines = f.readlines()
                for line in lines:
                    ex = json.loads(line)
                    context, question, answer = ex['context'], ex['question'], ex['answer']
                    examples.append(
                        Example.from_raw(make_example_id(self, len(examples)), context, question, answer, lower=lower)
                    )
                    if subsample is not None and len(examples) >= subsample:
                        break
            os.makedirs(os.path.dirname(cache_name), exist_ok=True)
            logger.info(f'Caching data to {cache_name}')
            torch.save(examples, cache_name)

        super(JSON, self).__init__(examples, **kwargs)

    @classmethod
    def splits(cls, root='.data', name=None, train='train', validation='val', test='test', **kwargs):
        path = os.path.join(root, name)

        train_data = None if train is None else cls(os.path.join(path, 'train.jsonl'), **kwargs)
        validation_data = None if validation is None else cls(os.path.join(path, 'val.jsonl'), **kwargs)
        test_data = None if test is None else cls(os.path.join(path, 'test.jsonl'), **kwargs)

        aux_data = None
        do_curriculum = kwargs.get('curriculum', False)
        if do_curriculum:
            kwargs.pop('curriculum')
            aux_data = cls(os.path.join(path, 'aux.jsonl'), **kwargs)

        return Split(
            train=None if train is None else train_data,
            eval=None if validation is None else validation_data,
            test=None if test is None else test_data,
            aux=None if do_curriculum is None else aux_data,
        )


class CrossNERDataset(CQA):
    is_classification = True

    def __init__(self, data, *, make_example, **kwargs):

        subsample = kwargs.pop('subsample')
        domain = kwargs.pop('domain')
        examples = []

        example_id, tokens, labels = 0, [], []
        for i, line in enumerate(data):
            line = line.strip()
            if line == "":
                # reached end of this example
                if len(tokens):
                    examples.append(make_example([example_id, tokens, labels], domain))
                tokens, labels = [], []
                example_id += 1
            else:
                splits = line.split("\t")
                tokens.append(splits[0])
                labels.append(splits[1])

            if subsample is not None and len(examples) >= subsample:
                break

        super().__init__(examples, **kwargs)

    @classmethod
    def return_splits(cls, name, path='.data', train='train', validation='dev', test='test', **kwargs):

        ner_domains = kwargs.pop('ner_domains')

        all_train_data = []
        all_validation_data = []
        all_test_data = []
        for domain in ner_domains:
            # download datasets and cache them
            train_data, validation_data, test_data = None, None, None
            train_path, validation_path, test_path = None, None, None
            if train:
                train_path = os.path.join(path, domain, 'train.txt')
                with open(train_path, "r") as fin:
                    train_data = fin.readlines()
            if validation:
                validation_path = os.path.join(path, domain, f'{validation}.txt')
                with open(validation_path, "r") as fin:
                    validation_data = fin.readlines()
            if test:
                test_path = os.path.join(path, domain, 'test.txt')
                with open(test_path, "r") as fin:
                    test_data = fin.readlines()

            # Uncomment for testing
            if kwargs.pop("hf_test_overfit", False):
                if validation:
                    validation_path = os.path.join(path, domain, 'train.txt')
                    with open(validation_path, "r") as fin:
                        validation_data = fin.readlines()
                if test:
                    test_path = os.path.join(path, domain, 'train.txt')
                    with open(test_path, "r") as fin:
                        test_data = fin.readlines()

            kwargs['domain'] = domain

            train_data = None if train is None else cls(train_data, **kwargs)
            validation_data = None if validation is None else cls(validation_data, **kwargs)
            test_data = None if test is None else cls(test_data, **kwargs)

            if not all_train_data:
                all_train_data = train_data
            elif train_data:
                all_train_data.examples = all_train_data.examples + train_data.examples
            if not all_validation_data:
                all_validation_data = validation_data
            elif validation_data:
                all_validation_data.examples = all_validation_data.examples + validation_data.examples
            if not all_test_data:
                all_test_data = test_data
            elif test_data:
                all_test_data.examples = all_test_data.examples + test_data.examples

        return Split(train=all_train_data, eval=all_validation_data, test=all_test_data), Split(
            train=train_path, eval=validation_path, test=test_path
        )



#
# Copyright (c) 2018, Salesforce, Inc.
#                     The Board of Trustees of the Leland Stanford Junior University
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import logging
import random

import torch

logger = logging.getLogger(__name__)

_warned_for_batch_size = False


class LengthSortedIterator(torch.utils.data.Sampler):
    """ """

    def __init__(self, data_source, batch_size, sort, shuffle_and_repeat, sort_key_fn, batch_size_fn, groups=1):
        """
        batch_size: can be number of tokens or number of examples, the type is inferred from batch_size_fn
        sort: if False, disables sorting and uses the original order. Useful for evaluation.
        shuffle_and_repeat: if True, the order of returned examples are semi-shuffled, and there is no end to the iterator
        groups: used for sentence batching
        """
        if groups is None:
            groups = 1
        assert batch_size % groups == 0
        assert len(data_source) % groups == 0

        self.sort_key = sort_key_fn
        self.batch_size_fn = batch_size_fn
        self.groups = groups

        if sort:
            # sort while keeping track of the original order
            data_with_original_order = list(
                zip(data_source, range(len(data_source)))
            )  # list of tuples of the form (data_source[i], i)
            # sort based on data_source
            sorted_data_with_original_order = sorted(
                data_with_original_order, key=lambda x: self.sort_key(x[0]), reverse=True
            )  # sort from long to short
            # separate the two parts of each tuple
            self.data_source, self.original_order = tuple(zip(*sorted_data_with_original_order))
        else:
            self.data_source, self.original_order = data_source, list(range(len(data_source)))
        self.batch_size = batch_size  # number of examples or number of tokens
        self.shuffle_and_repeat = shuffle_and_repeat
        self.last_batch_start_index = 0
        self.last_batch_start_index = self._get_next_batch_start_index()

        if not self.shuffle_and_repeat:
            # do not allow skipping examples during validation/ prediction
            self.no_skip = True
            # quickly iterate over self to calculate length
            self.length = 0
            for _ in self:
                self.length += 1
            # reset state
            self.last_batch_start_index = 0
            self.last_batch_start_index = self._get_next_batch_start_index()
        else:
            self.no_skip = False
            self.length = len(self.data_source)

    def __len__(self):
        return self.length

    def __iter__(self):
        self.last_batch_start_index = 0
        self.last_batch_start_index = self._get_next_batch_start_index()
        return self

    def __next__(self):
        batch_of_indices = []
        current_batch_size = 0
        candidate_index = self._get_next_batch_start_index()
        if candidate_index >= len(self.data_source):
            # This is the end of the iterator
            assert not self.shuffle_and_repeat
            raise StopIteration
        while current_batch_size < self.batch_size:
            candidate_example = self.data_source[candidate_index]
            if self.batch_size_fn([candidate_example]) > self.batch_size:
                global _warned_for_batch_size
                if self.no_skip:
                    raise ValueError(
                        'Have to skip examples in validation/ prediction splits. Increase the validation batch size'
                    )
                if not _warned_for_batch_size:
                    logger.warning(
                        'Skipping an example larger than batch size. Consider increasing the batch size to avoid this warning'
                    )
                    _warned_for_batch_size = True
                self.last_batch_start_index += 1
                candidate_index += 1
                if candidate_index >= len(self.data_source):
                    # This is the end of the iterator
                    assert not self.shuffle_and_repeat
                    raise StopIteration
                continue

            candidate_batch_size = self.batch_size_fn(
                [self.data_source[i] for i in batch_of_indices] + [candidate_example]
            )  # the new batch size if we added this example to the batch
            if candidate_batch_size > self.batch_size:
                # the new example would put us over the batch size limit
                break

            batch_of_indices.append(candidate_index)
            current_batch_size = candidate_batch_size
            candidate_index += 1

            if candidate_index == len(self.data_source):
                break  # don't start from i=0; there is a large difference between the length of the first and last element

        self.last_batch_start_index += len(batch_of_indices)
        return batch_of_indices

    def _get_next_batch_start_index(self):
        if self.shuffle_and_repeat:
            # if self.groups > 1, this ensures that the start of each batch is a multiply of self.groups, i.e. where a group starts
            return random.randrange(0, len(self.data_source) / self.groups) * self.groups
        else:
            return self.last_batch_start_index



#
# Copyright (c) 2020-2021 The Board of Trustees of the Leland Stanford Junior University
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


import logging
import time
from collections import defaultdict

import ujson
from elasticsearch import Elasticsearch, RequestsHttpConnection, exceptions
from elasticsearch.client import XPackClient
from elasticsearch.client.utils import NamespacedClient

from .database import is_banned
from .database_utils import has_overlap, normalize_text

ES_RETRY_ATTEMPTS = 5

tracer = logging.getLogger('elasticsearch')
tracer.setLevel(logging.CRITICAL)

logger = logging.getLogger(__name__)


class XPackClientMPCompatible(XPackClient):
    def __getattr__(self, attr_name):
        return NamespacedClient.__getattribute__(self, attr_name)


class RemoteElasticDatabase(object):
    def __init__(self, config, typeqid2id, ned_features_default_val, ned_features_size):
        self.typeqid2id = typeqid2id
        self.id2type = {v: k for k, v in self.typeqid2id.items()}
        self.unk_id = ned_features_default_val[0]
        self.unk_type = self.id2type[self.unk_id]

        self.ned_features_default_val = ned_features_default_val
        self.ned_features_size = ned_features_size

        self.canonical2type = defaultdict(list)
        self.auth = (config['username'], config['password'])
        self.index = config['index']
        self.host = config['host']
        self.es = Elasticsearch(
            hosts=[{'host': self.host, 'port': 443}],
            http_auth=self.auth,
            use_ssl=True,
            verify_certs=True,
            connection_class=RequestsHttpConnection,
            maxsize=8,
        )

        # use xpack client that is compatible with multiprocessing
        self.es.xpack = XPackClientMPCompatible(self.es)

    def lookup(self, tokens, min_entity_len=2, max_entity_len=4, answer_entities=None):

        if answer_entities is not None:
            tokens_type_ids = self.lookup_entities(tokens, answer_entities)
        else:
            tokens_type_ids = self.single_lookup(tokens, min_entity_len, max_entity_len)

        return tokens_type_ids

    def lookup_entities(self, tokens, entities, allow_fuzzy=False):
        tokens_type_ids = [[self.unk_id] * self.ned_features_size[0]] * len(tokens)
        tokens_text = " ".join(tokens)

        all_matches = self.batch_find_matches(entities, allow_fuzzy)

        for i in range(len(all_matches)):
            match = all_matches[i]
            ent = entities[i]
            assert len(match) in [0, 1]

            # sometimes we have tokens like '.' or ',' which receives no match
            if len(match) == 0:
                continue

            match = match[0]
            canonical = match['_source']['canonical']
            type = match['_source']['type']
            if not is_banned(canonical) and canonical == ent:
                type = self.typeqid2id.get(type, self.unk_id)

                ent_num_tokens = len(ent.split(' '))
                idx = tokens_text.index(ent)
                token_pos = len(tokens_text[:idx].strip().split(' '))

                tokens_type_ids[token_pos : token_pos + ent_num_tokens] = [[type] * self.ned_features_size[0]] * ent_num_tokens

        return tokens_type_ids

    def batch_find_matches(self, keys_list, allow_fuzzy):

        if allow_fuzzy:
            query_temp = ujson.dumps(
                {
                    "size": 1,
                    "query": {"multi_match": {"query": "{}", "fields": ["canonical^8", "aliases^3"], "fuzziness": "AUTO"}},
                }
            )
        else:
            query_temp = ujson.dumps(
                {"size": 1, "query": {"multi_match": {"query": "{}", "fields": ["canonical^8", "aliases^3"]}}}
            )

        queries = []
        for key in keys_list:
            queries.append(ujson.loads(query_temp.replace('"{}"', '"' + key + '"')))

        search_header = ujson.dumps({'index': self.index})
        request = ''
        for q in queries:
            request += '{}\n{}\n'.format(search_header, ujson.dumps(q))

        retries = 0
        result = None
        while retries < ES_RETRY_ATTEMPTS:
            try:
                result = self.es.msearch(index=self.index, body=request)
                break
            except (exceptions.ConnectionTimeout, exceptions.ConnectionError, exceptions.TransportError):
                logger.warning('Connection Timed Out!')
                time.sleep(1)
                retries += 1

        if not result:
            raise ConnectionError('Could not reconnect to ES database after {} tries...'.format(ES_RETRY_ATTEMPTS))

        matches = [res['hits']['hits'] for res in result['responses']]
        return matches

    def single_lookup(self, tokens, min_entity_len, max_entity_len, allow_fuzzy=False):
        # load nltk lazily
        import nltk

        nltk.download('averaged_perceptron_tagger', quiet=True)

        tokens_type_ids = [[self.typeqid2id['unk']] * self.ned_features_size[0]] * len(tokens)

        max_entity_len = min(max_entity_len, len(tokens))
        min_entity_len = min(min_entity_len, len(tokens))

        pos_tagged = nltk.pos_tag(tokens)
        verbs = set([x[0] for x in pos_tagged if x[1].startswith('V')])

        used_aliases = []
        gram_attempts = []
        for n in range(max_entity_len, min_entity_len - 1, -1):
            ngrams = nltk.ngrams(tokens, n)
            start = -1
            end = n - 1
            for gram in ngrams:
                start += 1
                end += 1
                gram_text = normalize_text(" ".join(gram))

                if not is_banned(gram_text) and gram_text not in verbs:
                    if has_overlap(start, end, used_aliases):
                        continue
                    gram_attempts.append(gram_text)

            all_matches = self.batch_find_matches(gram_attempts, allow_fuzzy)
            for i in range(len(all_matches)):
                match = all_matches[i]
                assert len(match) in [0, 1]

                # sometimes we have tokens like '.' or ',' which receives no match
                if len(match) == 0:
                    continue

                match = match[0]
                canonical = match['_source']['canonical']
                type = match['_source']['type']
                if not is_banned(canonical) and canonical == gram_attempts[i]:
                    used_aliases.append([self.typeqid2id.get(type, self.unk_id), start, end])

        for type_id, beg, end in used_aliases:
            tokens_type_ids[beg:end] = [[type_id] * self.ned_features_size[0]] * (end - beg)

        return tokens_type_ids

    def batch_lookup(self, tokens_list, allow_fuzzy):
        length = len(tokens_list)

        all_currs = [0] * length
        all_lengths = [len(tokens) for tokens in tokens_list]
        all_ends = [len(tokens) for tokens in tokens_list]
        all_tokens_type_ids = [[] for _ in range(length)]
        all_done = [False] * length

        while not all(all_done):
            batch_to_lookup = [' '.join(tokens[all_currs[i] : all_ends[i]]) for i, tokens in enumerate(tokens_list)]

            all_matches = self.batch_find_matches(batch_to_lookup, allow_fuzzy)

            for i in range(length):
                if all_done[i]:
                    continue

                match = all_matches[i]
                assert len(match) in [0, 1]

                # sometimes we have tokens like '.' or ',' which receives no match
                if len(match) == 0:
                    all_ends[i] -= 1
                    continue

                match = match[0]
                canonical = match['_source']['canonical']
                type = match['_source']['type']
                if not is_banned(canonical) and canonical == batch_to_lookup[i]:
                    all_tokens_type_ids[i].extend([self.typeqid2id[type] for _ in range(all_currs[i], all_ends[i])])
                    all_currs[i] = all_ends[i]
                else:
                    all_ends[i] -= 1

            for j in range(length):
                # no matches were found for the span starting from current index
                if all_currs[j] == all_ends[j] and not all_currs[j] >= all_lengths[j]:
                    all_tokens_type_ids[j].append(self.typeqid2id['unk'])
                    all_currs[j] += 1
                    all_ends[j] = all_lengths[j]

            for j in range(length):
                # reached end of sentence
                if all_currs[j] >= all_lengths[j]:
                    all_done[j] = True

        return all_tokens_type_ids

    def es_dump(self, mode):

        assert mode in ['canonical2type', 'typeqid2id'], 'Mode not supported; use either canonical2type or typeqid2id'

        if mode == 'canonical2type':
            return_dict = defaultdict(list)
            dump_file_name = 'canonical2type'
        elif mode == 'typeqid2id':
            return_dict = defaultdict(int)
            dump_file_name = 'typeqid2id'

        begin = time.time()

        def do_search(scroll_id, total_values):
            body = {"size": 10000, "query": {"match_all": {}}}

            if scroll_id is None:
                result = self.es.search(index=self.index, body=body, scroll='3m')
            else:
                result = self.es.scroll(scroll_id=scroll_id, scroll='3m')
            total_values += len(result["hits"]["hits"])
            if len(result["hits"]["hits"]) < 10:
                return None, -1
            print("total docs:", total_values)
            for match in result["hits"]["hits"]:
                if mode == 'canonical2type':
                    return_dict[match['_source']['canonical']].append(match['_source']['type'][len('org.wikidata:') :])
                elif mode == 'typeqid2id':
                    if match['_source']['type'] not in return_dict:
                        return_dict[match['_source']['type']] = len(self.typeqid2id)

            scroll_id = result['_scroll_id']
            print('processed: {}, time elapsed: {}'.format(i, time.time() - begin))

            return scroll_id, total_values

        scroll_id, total_values = do_search(None, 0)
        i = 0
        chunk = 0
        while True:
            if total_values % 4000000 == 0:
                with open(f'{dump_file_name}_{chunk}.json', 'w') as fout:
                    ujson.dump(return_dict, fout, ensure_ascii=True)
                chunk += 1
                return_dict.clear()
            try:
                scroll_id, total_values = do_search(scroll_id, total_values)
                if scroll_id is None and total_values == -1:
                    break
                i += 1
            except BaseException:
                break

        # dump any remaining values
        with open(f'{dump_file_name}_{chunk}.json', 'w') as fout:
            ujson.dump(return_dict, fout, ensure_ascii=True)



#
# Copyright (c) 2020-2021 The Board of Trustees of the Leland Stanford Junior University
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


import torch
from sentence_transformers import SentenceTransformer
from sklearn.metrics.pairwise import paired_cosine_distances


def parse_argv(parser):

    parser.add_argument('--input_file', type=str)
    parser.add_argument('--output_file', type=str)
    parser.add_argument('--batch_size', type=int, default=250)
    parser.add_argument('--subsample', type=int, default=-1)
    parser.add_argument('--device', type=str, default='cuda')
    parser.add_argument(
        '--model_name',
        type=str,
        default='xlm-r-distilroberta-base-paraphrase-v1',
        help='List of available models: https://public.ukp.informatik.tu-darmstadt.de/reimers/sentence-transformers/v0.2/',
    )


def main(args):

    model = SentenceTransformer(args.model_name)

    if args.device == 'cuda' and torch.cuda.is_available():
        model.cuda()

    ids = []
    src_sentences = []
    trg_sentences = []
    programs = []

    with open(args.input_file, 'r') as fin:
        for i, line in enumerate(fin):
            row = list(map(lambda part: part.strip(), line.split('\t')))
            ids.append(row[0])
            src_sentences.append(row[1])
            trg_sentences.append(row[2])
            programs.append(row[3])

            if args.subsample != -1 and i >= args.subsample:
                break

    embeddings1 = model.encode(src_sentences, batch_size=args.batch_size, show_progress_bar=True, convert_to_numpy=True)
    embeddings2 = model.encode(trg_sentences, batch_size=args.batch_size, show_progress_bar=True, convert_to_numpy=True)

    cosine_scores = 1 - (paired_cosine_distances(embeddings1, embeddings2))

    with open(args.output_file, 'w') as fout:
        for (
            id_,
            src,
            tgt,
            score,
            program,
        ) in zip(ids, src_sentences, trg_sentences, cosine_scores, programs):
            fout.write('\t'.join([id_, src, tgt, '{:0.4f}'.format(score), program]) + '\n')



#
# Copyright (c) 2020 The Board of Trustees of the Leland Stanford Junior University
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import math
import sys

from tqdm import tqdm


class LogFriendlyProgressBar:
    def __init__(self, iterable, desc, total, step: int = 1):
        """
        Inputs:
            step: Will print progress every `step` %
        """
        self._desc = desc
        self._i = 0
        self._N = total
        self.step = step
        self._progress = 0
        self._iterable = iterable
        self._iterator = None

    def __iter__(self):
        self._iterator = iter(self._iterable)
        return self

    def __next__(self):
        value = next(self._iterator)
        self._i += 1
        progress = math.floor(self._i * 100 / self._N)
        if progress > self._progress and progress % self.step == 0:
            if self._desc:
                print(f'{self._desc} - progress: {progress}%', file=sys.stderr)
            else:
                print(f'Progress: {progress}%', file=sys.stderr)
            sys.stderr.flush()
            self._progress = progress
        return value

    def close(self):
        return


def progress_bar(iterable, desc=None, total=None, disable=False):
    if disable:
        return iterable
    if total is None:
        if not hasattr(iterable, '__len__'):
            return iterable
        total = len(iterable)
    if sys.stderr.isatty():
        return tqdm(iterable, desc=desc, total=total)
    else:
        return LogFriendlyProgressBar(iterable, desc=desc, total=total, step=5)


def prange(*args, desc=None, disable=False):
    return progress_bar(range(*args), desc=desc, disable=disable)



#
# Copyright (c) 2020-2021 The Board of Trustees of the Leland Stanford Junior University
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import logging

from .database_utils import has_overlap, is_banned, normalize_text

tracer = logging.getLogger('elasticsearch')
tracer.setLevel(logging.CRITICAL)

logger = logging.getLogger(__name__)


class Database(object):
    def __init__(self, canonical2type, typeqid2id, all_canonicals, ned_features_default_val, ned_features_size):
        self.canonical2type = canonical2type
        self.typeqid2id = typeqid2id
        self.id2type = {v: k for k, v in self.typeqid2id.items()}
        self.all_canonicals = all_canonicals

        self.unk_id = ned_features_default_val[0]
        self.unk_type = self.id2type[self.unk_id]

        self.ned_features_default_val = ned_features_default_val
        self.ned_features_size = ned_features_size

    def lookup_ngrams(self, tokens, min_entity_len, max_entity_len):
        # load nltk lazily
        import nltk

        nltk.download('averaged_perceptron_tagger', quiet=True)

        tokens_type_ids = [[self.unk_id] * self.ned_features_size[0]] * len(tokens)

        max_entity_len = min(max_entity_len, len(tokens))
        min_entity_len = min(min_entity_len, len(tokens))

        pos_tagged = nltk.pos_tag(tokens)
        verbs = set([x[0] for x in pos_tagged if x[1].startswith('V')])

        used_aliases = []
        for n in range(max_entity_len, min_entity_len - 1, -1):
            ngrams = nltk.ngrams(tokens, n)
            start = -1
            end = n - 1
            for gram in ngrams:
                start += 1
                end += 1
                gram_text = normalize_text(" ".join(gram))

                if not is_banned(gram_text) and gram_text not in verbs and gram_text in self.all_canonicals:
                    if has_overlap(start, end, used_aliases):
                        continue

                    used_aliases.append([self.typeqid2id.get(self.canonical2type[gram_text], self.unk_id), start, end])

        for type_id, beg, end in used_aliases:
            tokens_type_ids[beg:end] = [[type_id] * self.ned_features_size[0]] * (end - beg)

        return tokens_type_ids

    def lookup_smaller(self, tokens):

        tokens_type_ids = []
        i = 0
        while i < len(tokens):
            token = tokens[i]
            # sort by number of tokens so longer keys get matched first
            matched_items = sorted(self.all_canonicals.keys(token), key=lambda item: len(item), reverse=True)
            found = False
            for key in matched_items:
                type = self.canonical2type[key]
                key_tokenized = key.split()
                cur = i
                j = 0
                while cur < len(tokens) and j < len(key_tokenized):
                    if tokens[cur] != key_tokenized[j]:
                        break
                    j += 1
                    cur += 1

                if j == len(key_tokenized):
                    if is_banned(' '.join(key_tokenized)):
                        continue

                    # match found
                    found = True
                    tokens_type_ids.extend([[self.typeqid2id[type] * self.ned_features_size[0]] for _ in range(i, cur)])

                    # move i to current unprocessed position
                    i = cur
                    break

            if not found:
                tokens_type_ids.append([self.unk_id * self.ned_features_size[0]])
                i += 1

        return tokens_type_ids

    def lookup_longer(self, tokens):
        i = 0
        tokens_type_ids = []

        length = len(tokens)
        found = False
        while i < length:
            end = length
            while end > i:
                tokens_str = ' '.join(tokens[i:end])
                if tokens_str in self.all_canonicals:
                    # match found
                    found = True
                    tokens_type_ids.extend(
                        [[self.typeqid2id[self.canonical2type[tokens_str]] * self.ned_features_size[0]] for _ in range(i, end)]
                    )
                    # move i to current unprocessed position
                    i = end
                    break
                else:
                    end -= 1
            if not found:
                tokens_type_ids.append([self.unk_id * self.ned_features_size[0]])
                i += 1
            found = False

        return tokens_type_ids

    def lookup_entities(self, tokens, entities):
        tokens_type_ids = [[self.unk_id] * self.ned_features_size[0]] * len(tokens)
        tokens_text = " ".join(tokens)

        for ent in entities:
            if ent not in self.all_canonicals:
                continue
            ent_num_tokens = len(ent.split(' '))
            idx = tokens_text.index(ent)
            token_pos = len(tokens_text[:idx].strip().split(' '))

            type = self.typeqid2id.get(self.canonical2type[ent], self.unk_id)

            tokens_type_ids[token_pos : token_pos + ent_num_tokens] = [[type] * self.ned_features_size[0]] * ent_num_tokens

        return tokens_type_ids

    def lookup(self, tokens, database_lookup_method=None, min_entity_len=2, max_entity_len=4, answer_entities=None):

        tokens_type_ids = [[self.unk_id] * self.ned_features_size[0]] * len(tokens)

        if answer_entities is not None:
            tokens_type_ids = self.lookup_entities(tokens, answer_entities)
        else:
            if database_lookup_method == 'smaller_first':
                tokens_type_ids = self.lookup_smaller(tokens)
            elif database_lookup_method == 'longer_first':
                tokens_type_ids = self.lookup_longer(tokens)
            elif database_lookup_method == 'ngrams':
                tokens_type_ids = self.lookup_ngrams(tokens, min_entity_len, max_entity_len)

        return tokens_type_ids



#
# Copyright (c) 2018, Salesforce, Inc.
#                     The Board of Trustees of the Leland Stanford Junior University
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import logging

import torch
from transformers import AutoConfig, AutoModel, BertConfig, PretrainedConfig, XLMRobertaConfig

from ..data_utils.numericalizer import TransformerNumericalizer
from ..paraphrase.transformers_utils import BertModelForNER, XLMRobertaModelForNER
from ..util import adjust_language_code
from .base import GenieModel
from .identity_encoder import IdentityEncoder
from .mqan_decoder import MQANDecoder

logger = logging.getLogger(__name__)


class TransformerLSTM(GenieModel):
    def __init__(self, config=None, *inputs, args, vocab_sets, tasks, save_directory=None, **kwargs):
        """
        Relevant inputs should be provided using kwargs. This method is defined this way to match parent's and siblings' method signatures.
        If `save_directory` is None, will initialize a new model and numericalizer, otherwise, will load them from `save_directory`
        Inputs:
            args
            vocab_sets
            save_directory: The directory where numericalizer can be loaded from.
        """
        super().__init__(PretrainedConfig())  # dummy PretrainedConfig
        self.args = args

        encoder_embeddings = args.pretrained_model
        config = AutoConfig.from_pretrained(encoder_embeddings, cache_dir=args.embeddings)
        args.dimension = config.hidden_size

        self.src_lang, self.tgt_lang = adjust_language_code(
            config, args.pretrained_model, kwargs.get('src_lang', 'en'), kwargs.get('tgt_lang', 'en')
        )

        self.numericalizer = TransformerNumericalizer(
            encoder_embeddings,
            args,
            max_generative_vocab=args.max_generative_vocab,
            save_dir=save_directory,
            config=config,
            src_lang=self.src_lang,
            tgt_lang=self.tgt_lang,
            vocab_sets=vocab_sets,
            tasks=tasks,
        )

        logger.info('Initializing encoder and decoder embeddings')

        if args.do_ned:
            if type(config) == BertConfig:
                if save_directory is not None:
                    self.encoder_embeddings = BertModelForNER(config, args.num_db_types, args.db_unk_id)
                else:
                    self.encoder_embeddings = BertModelForNER(config, args.num_db_types, args.db_unk_id).from_pretrained(
                        encoder_embeddings, num_db_types=args.num_db_types, db_unk_id=args.db_unk_id, cache_dir=args.embeddings
                    )
            elif type(config) == XLMRobertaConfig:
                if save_directory is not None:
                    self.encoder_embeddings = XLMRobertaModelForNER(config, args.num_db_types, args.db_unk_id)
                else:
                    self.encoder_embeddings = XLMRobertaModelForNER(config, args.num_db_types, args.db_unk_id).from_pretrained(
                        encoder_embeddings, num_db_types=args.num_db_types, db_unk_id=args.db_unk_id, cache_dir=args.embeddings
                    )
            else:
                raise ValueError('Model is not supported for using entity embeddings for NER')
        else:
            if save_directory is not None:
                self.encoder_embeddings = AutoModel.from_config(config)
            else:
                self.encoder_embeddings = AutoModel.from_pretrained(
                    encoder_embeddings, config=config, cache_dir=args.embeddings
                )

        self.encoder_embeddings.resize_token_embeddings(self.numericalizer.num_tokens)

        logger.info(f'Vocabulary has {self.numericalizer.num_tokens} tokens')

        self.encoder = IdentityEncoder(self.numericalizer, args, config, self.encoder_embeddings)
        self.decoder = MQANDecoder(self.numericalizer, args)

    def add_new_vocab_from_data(self, tasks, resize_decoder=False):
        super().add_new_vocab_from_data(tasks, resize_decoder=resize_decoder)
        self.encoder_embeddings.resize_token_embeddings(self.numericalizer.num_tokens)
        if resize_decoder:
            self.decoder.decoder_embeddings.resize_embedding(self.numericalizer.num_tokens)

    # return_dict, output_scores, output_attentions, output_hidden_states are unused but needed since the base class (PreTrainedModel) passes them
    def forward(
        self,
        batch,
        current_token_id=None,
        past_key_values=None,
        expansion_factor=1,
        generation_dict=None,
        encoder_output=None,
        return_dict=False,
        output_scores=False,
        output_attentions=False,
        output_hidden_states=False,
    ):
        if encoder_output is None:
            final_context, context_rnn_state = self.encoder(batch)
        else:
            final_context, context_rnn_state = encoder_output
        encoder_loss = None
        if self.training and getattr(self.args, 'use_encoder_loss', None):
            encoder_loss = self.get_encoder_loss(context_rnn_state)

        return self.decoder(
            batch,
            final_context,
            context_rnn_state,
            encoder_loss,
            current_token_id,
            decoder_wrapper=past_key_values,
            expansion_factor=expansion_factor,
            generation_dict=generation_dict,
        )

    def get_encoder_loss(self, context_rnn_state):

        # concat hidden and cell state
        if len(context_rnn_state) == 2:
            context_rnn_state = torch.cat(context_rnn_state, dim=0)

        batch_size = context_rnn_state.size(1)
        groups = len(self.args.train_src_languages.split('+'))
        assert batch_size % groups == 0

        # reshape to be (batch_size; -1)
        context_rnn_state = context_rnn_state.view(batch_size, -1)

        if self.args.encoder_loss_type == 'mean':
            # element-wise mean of encoder loss https://www.aclweb.org/anthology/W18-3023.pdf
            context_value = torch.mean(context_rnn_state, dim=-1)
        elif self.args.encoder_loss_type == 'sum':
            context_value = torch.sum(context_rnn_state, dim=-1)

        encoder_loss = 0.0
        for i in range(0, batch_size, groups):
            indices = [j for j in range(i, i + groups)]
            groups_vals = context_value[indices]
            assert len(groups_vals) > 1
            encoder_loss += torch.std(groups_vals).item()

        return encoder_loss

    def get_output_embeddings(self):
        return self.decoder.decoder_embeddings

    def prepare_inputs_for_generation(
        self, input_ids, attention_mask, use_cache, batch, generation_dict, encoder_output, past=None
    ):
        expansion_factor = input_ids.shape[0] // len(batch.example_id)
        return {
            "batch": batch,
            "past_key_values": past,
            "current_token_id": input_ids[:, -1:],
            "expansion_factor": expansion_factor,
            "generation_dict": generation_dict,
            "encoder_output": encoder_output,
        }

    def _reorder_cache(self, past, beam_idx):
        past.reorder(beam_idx)
        return past

    def generate(
        self,
        batch,
        max_output_length,
        num_outputs,
        temperature,
        repetition_penalty,
        top_k,
        top_p,
        num_beams,
        num_beam_groups,
        diversity_penalty,
        no_repeat_ngram_size,
        do_sample,
    ):

        encoder_output = self.encoder(batch)
        self.config.vocab_size = len(self.numericalizer.decoder_vocab)
        self.config.is_encoder_decoder = (
            False  # in order to make it work with `transformers` generation code, we should treat this as a decoder-only model
        )
        batch_size = len(batch.example_id)
        input_ids = torch.full((batch_size, 1), self.decoder.init_idx, dtype=torch.long, device=batch.context.value.device)

        generated = super().generate(
            input_ids=input_ids,
            batch=batch,
            max_length=max_output_length,
            min_length=2,  # generate at least one token after BOS
            bos_token_id=self.decoder.init_idx,
            pad_token_id=self.numericalizer.decoder_vocab.pad_idx,
            early_stopping=True,
            num_return_sequences=num_outputs,
            repetition_penalty=repetition_penalty,
            temperature=temperature,
            eos_token_id=self.numericalizer.decoder_vocab.eos_idx,
            top_k=top_k,
            top_p=top_p,
            num_beams=num_beams,
            num_beam_groups=num_beam_groups,
            diversity_penalty=diversity_penalty,
            no_repeat_ngram_size=no_repeat_ngram_size,
            do_sample=do_sample,
            generation_dict={'max_output_length': max_output_length},
            encoder_output=encoder_output,
            output_scores=False,
            output_attentions=False,
            output_hidden_states=False,
            return_dict_in_generate=True,
        )
        output_ids = generated.sequences
        mapped_output_ids = torch.cat(
            (output_ids[:, 0:1], output_ids[:, 1:].cpu().apply_(self.decoder.map_to_full).to(batch.context.value.device)),
            dim=1,
        )  # map everything to full vocabulary except BOS which already is in full vocabulary
        generated.sequences = mapped_output_ids

        return generated



#
# Copyright (c) 2020 The Board of Trustees of the Leland Stanford Junior University
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import logging
from typing import List

import torch
from torch.tensor import Tensor
from transformers import AutoConfig, AutoModelForSeq2SeqLM, MBartTokenizer, MBartTokenizerFast

from ..data_utils.numericalizer import TransformerNumericalizer
from ..util import ConfidenceFeatures, adjust_language_code
from .base import GenieModel
from .common import LabelSmoothingCrossEntropy

logger = logging.getLogger(__name__)


class TransformerSeq2Seq(GenieModel):
    def __init__(self, config=None, *inputs, args, tasks, vocab_sets, save_directory=None, **kwargs):
        """
        If `save_directory` is None, will initialize a new model and numericalizer, otherwise, will load them from `save_directory`
        """
        config = AutoConfig.from_pretrained(args.pretrained_model, cache_dir=args.embeddings)
        super().__init__(config)
        self.args = args
        args.dimension = config.d_model
        self._is_bart_large = self.args.pretrained_model == 'facebook/bart-large'
        self._is_mbart = 'mbart' in self.args.pretrained_model
        self._is_mbart50 = self._is_mbart and '-50-' in self.args.pretrained_model

        self.src_lang, self.tgt_lang = adjust_language_code(
            config, args.pretrained_model, kwargs.get('src_lang', 'en'), kwargs.get('tgt_lang', 'en')
        )

        if save_directory is not None:
            self.model = AutoModelForSeq2SeqLM.from_config(config)
        else:
            self.model = AutoModelForSeq2SeqLM.from_pretrained(self.args.pretrained_model, cache_dir=self.args.embeddings)

        self.numericalizer = TransformerNumericalizer(
            self.args.pretrained_model,
            args,
            max_generative_vocab=None,
            save_dir=save_directory,
            config=config,
            src_lang=self.src_lang,
            tgt_lang=self.tgt_lang,
            vocab_sets=vocab_sets,
            tasks=tasks,
        )

        self.model.resize_token_embeddings(self.numericalizer.num_tokens)

        # set decoder_start_token_id
        # recommended by huggingface
        # TODO check if it's actually useful
        if self.model.config.decoder_start_token_id is None and isinstance(
            self.numericalizer._tokenizer, (MBartTokenizer, MBartTokenizerFast)
        ):
            if isinstance(self.numericalizer._tokenizer, MBartTokenizer):
                self.model.config.decoder_start_token_id = self.numericalizer._tokenizer.lang_code_to_id[self.tgt_lang]
            else:
                self.model.config.decoder_start_token_id = self.numericalizer._tokenizer.convert_tokens_to_ids(self.tgt_lang)

        if self.model.config.decoder_start_token_id is None:
            raise ValueError("Make sure that decoder_start_token_id for the model is defined")

        if args.dropper_ratio > 0:
            # lazy import since dropper is an optional dependency
            from loss_dropper import LossDropper

            self.dropper = LossDropper(dropc=args.dropper_ratio, min_count=args.dropper_min_count)
        else:
            self.dropper = None

        self.criterion = LabelSmoothingCrossEntropy(args.label_smoothing)

    def add_new_vocab_from_data(self, tasks, resize_decoder=False):
        super().add_new_vocab_from_data(tasks, resize_decoder)
        self.model.resize_token_embeddings(self.numericalizer.num_tokens)

    def forward(self, *input, **kwargs):
        if self.training:
            batch = input[0]

            answer = batch.answer.value
            answer_length = batch.answer.length
            if self._is_bart_large:
                # remove BOS from the answer to BART-Large because BART-Large was not trained to predict BOS
                # (unlike BART-Base or mBART)
                #
                # NOTE: various people at Huggingface and elsewhere have tried to conclusively ascertain
                # whether BOS should be there or not, and the answer seems to be that BOS should not be there
                # at all, either in input or in the output
                # but empirically, BOS in the input works slightly better, perhaps because our sentences start
                # with a lowercase letter, so we leave it
                answer = answer[:, 1:].contiguous()
                answer_length = answer_length - 1

            # this has several differences compared to what `transformers` Seq2Seq models do:
            # (1) pad tokens are ignored in all loss calculations
            # (2) loss is averaged over sequence lengths first, then over the batch size. This way,
            # longer sequences in the batch do not drown shorter sequences.
            # (3) if `args.dropper_ratio > 0.0`, will perform Loss Truncation
            # (4) if `args.label_smoothing > 0.0`, will add label smoothing term to loss
            outputs = self.model(
                batch.context.value, labels=answer, attention_mask=(batch.context.value != self.numericalizer.pad_id)
            )
            batch_size, vocab_size = outputs.logits.shape[0], outputs.logits.shape[2]
            loss = self.criterion(
                outputs.logits.view(-1, vocab_size), target=answer.view(-1), ignore_index=self.numericalizer.pad_id
            )
            loss = loss.view(batch_size, -1)  # (batch_size, sequence_length)
            loss = loss.sum(dim=1) / answer_length  # accounts for the case where BOS is removed
            if self.dropper is not None:
                dropper_mask = self.dropper(loss)
                loss = loss * dropper_mask
            loss = loss.mean()  # average over the batch size
            outputs.loss = loss  # replace the loss calculated by `transformers` with the new loss
            return outputs
        else:
            return self.model(**kwargs)

    def generate(
        self,
        batch,
        max_output_length,
        num_outputs,
        temperature,
        repetition_penalty,
        top_k,
        top_p,
        num_beams,
        num_beam_groups,
        diversity_penalty,
        no_repeat_ngram_size,
        do_sample,
    ):

        decoder_start_token_id, forced_bos_token_id = None, None
        if self._is_mbart:
            decoder_start_token_id = self.model.config.decoder_start_token_id
        if self._is_mbart50:
            forced_bos_token_id = self.numericalizer._tokenizer.lang_code_to_id[self.tgt_lang]

        input_ids = batch.context.value
        # when attention_mask is not provided to generate(), it will default to masking pad tokens, which is the correct thing
        generated = self.model.generate(
            input_ids=input_ids,
            max_length=max_output_length,
            min_length=3,  # generate at least one token after BOS and language code
            bos_token_id=self.numericalizer.init_id,
            pad_token_id=self.numericalizer.pad_id,
            early_stopping=False,
            num_return_sequences=num_outputs,
            repetition_penalty=repetition_penalty,
            temperature=temperature,
            eos_token_id=self.numericalizer.eos_id,
            top_k=top_k,
            top_p=top_p,
            num_beams=num_beams,
            num_beam_groups=num_beam_groups,
            diversity_penalty=diversity_penalty,
            no_repeat_ngram_size=no_repeat_ngram_size,
            do_sample=do_sample,
            decoder_start_token_id=decoder_start_token_id,
            forced_bos_token_id=forced_bos_token_id,
            output_scores=False,
            output_attentions=True,
            output_hidden_states=False,
            return_dict_in_generate=True,
        )

        return generated

    def confidence_features(self, batch, predictions, mc_dropout_num=0) -> List[ConfidenceFeatures]:
        """
        predictions: Tensor of shape (batch_size, output_length)
        mc_droput_num: number of Monte Carlo samples used for the MC Dropout method. 0 disables MC dropout.
        """
        batch_size = predictions.shape[0]
        repetition_factor = batch_size // batch.context.value.shape[0]
        input_ids = batch.context.value.repeat_interleave(
            repetition_factor, dim=0
        )  # repeat to account for multiple predictions per input

        prediction_lengths = self.get_length(predictions)

        pad_token_id = self.numericalizer.pad_id
        attention_mask = self.model._prepare_attention_mask_for_generation(
            input_ids=input_ids, pad_token_id=pad_token_id, eos_token_id=self.numericalizer.eos_id
        )
        truncated_predictions = predictions[:, 1:]  # remove the BOS token since it is not actually being generated

        assert not self.training, 'Model should be in eval() mode before generation can start.'

        batch_nodrop_logits = []
        batch_nodrop_probs = []
        batch_nodrop_entropies = []
        # batch_nodrop_top1_probs = []
        # batch_nodrop_top1_idx = []
        # batch_nodrop_top2_probs = []
        # batch_nodrop_top2_idx = []
        outputs = self.model(
            input_ids=input_ids,
            decoder_input_ids=predictions,
            attention_mask=attention_mask,
            return_dict=True,
            use_cache=False,
        )
        nodrop_logits = outputs.logits[:, :-1, :]  # remove the last probability distribution which is for the token after EOS
        for i in range(batch_size):
            batch_nodrop_logits.append(
                nodrop_logits[i].gather(dim=1, index=truncated_predictions[i].view(-1, 1)).view(-1)[: prediction_lengths[i]]
            )
            probs = torch.softmax(nodrop_logits[i], dim=1)
            batch_nodrop_probs.append(
                probs.gather(dim=1, index=truncated_predictions[i].view(-1, 1)).view(-1)[: prediction_lengths[i]]
            )
            batch_nodrop_entropies.append(-torch.sum(torch.log(probs) * probs, dim=1)[: prediction_lengths[i]])
            # sorted_probs = probs.sort(dim=1)
            # batch_nodrop_top1_probs.append(sorted_probs.values[:, -1][:prediction_lengths[i]])
            # batch_nodrop_top2_probs.append(sorted_probs.values[:, -2][:prediction_lengths[i]])
            # batch_nodrop_top1_idx.append(sorted_probs.indices[:, -1][:prediction_lengths[i]])
            # batch_nodrop_top2_idx.append(sorted_probs.indices[:, -2][:prediction_lengths[i]])

        # activate dropout layers
        self.train()

        batch_drop_logits = [[] for _ in range(batch_size)]
        batch_drop_probs = [[] for _ in range(batch_size)]
        # batch_drop_top1_probs = [[] for _ in range(batch_size)]
        # batch_drop_top2_probs = [[] for _ in range(batch_size)]
        for _ in range(mc_dropout_num):
            outputs = self.model(
                input_ids=input_ids,
                decoder_input_ids=predictions,
                attention_mask=attention_mask,
                return_dict=True,
                use_cache=False,
            )
            drop_logits = outputs.logits[
                :, :-1, :
            ]  # remove the last probability distribution which is for the token after EOS
            for i in range(batch_size):
                batch_drop_logits[i].append(
                    (drop_logits[i].gather(dim=1, index=truncated_predictions[i].view(-1, 1)).view(-1))[
                        : prediction_lengths[i]
                    ]
                )
                softmax = torch.softmax(drop_logits[i], dim=1)
                drop_probs = softmax.gather(dim=1, index=truncated_predictions[i].view(-1, 1)).view(-1)[
                    : prediction_lengths[i]
                ]
                batch_drop_probs[i].append(drop_probs)
                # batch_drop_top1_probs[i].append(softmax.gather(dim=1, index=batch_nodrop_top1_idx[i].view(-1, 1)).view(-1)[:prediction_lengths[i]])
                # batch_drop_top2_probs[i].append(softmax.gather(dim=1, index=batch_nodrop_top2_idx[i].view(-1, 1)).view(-1)[:prediction_lengths[i]])

        confidence_features = []
        for i in range(batch_size):
            # TODO refactor so token changes (EOS/ BOS removal, etc.) is decided based on specific tokenizer used
            if self._is_bart_large:
                prediction = predictions[i][: prediction_lengths[i] + 1]  # +1 to include EOS
            else:
                prediction = predictions[i][1 : prediction_lengths[i] + 1]  # remove token before BOS, +1 to include EOS
            confidence_features.append(
                ConfidenceFeatures(
                    drop_logits=batch_drop_logits[i] if mc_dropout_num > 0 else None,
                    drop_probs=batch_drop_probs[i] if mc_dropout_num > 0 else None,
                    #  drop_top1_probs=batch_drop_top1_probs[i] if mc_dropout_num > 0 else None,
                    #  drop_top2_probs=batch_drop_top2_probs[i] if mc_dropout_num > 0 else None,
                    gold_answer=batch.answer.value[i // repetition_factor][: batch.answer.length[i // repetition_factor]],
                    prediction=prediction,
                    nodrop_logits=batch_nodrop_logits[i][: prediction_lengths[i]],
                    nodrop_probs=batch_nodrop_probs[i][: prediction_lengths[i]],
                    #  nodrop_top1_probs=batch_nodrop_top1_probs[i][:prediction_lengths[i]],
                    #  nodrop_top2_probs=batch_nodrop_top2_probs[i][:prediction_lengths[i]],
                    nodrop_entropies=batch_nodrop_entropies[i][: prediction_lengths[i]],
                    context=batch.context.value[i // repetition_factor][: batch.context.length[i // repetition_factor]],
                )
            )

        # return the model back to its previous state
        self.eval()

        return confidence_features

    def get_length(self, prediction: Tensor):
        # skip the first token, because BOS is the same as EOS for some models
        prediction = prediction[:, 1:]

        # add EOS at the end in case the prediction doesn't have any
        prediction = torch.cat(
            [
                prediction,
                torch.ones((prediction.shape[0], 1), dtype=torch.long, device=prediction.device) * self.numericalizer.eos_id,
            ],
            dim=1,
        )

        # find the index of the first eos
        first_eos_one_hot = (torch.cumsum((prediction == self.numericalizer.eos_id).long(), dim=1) == 1) & (
            prediction == self.numericalizer.eos_id
        )
        first_eos = first_eos_one_hot.nonzero(as_tuple=False)[:, 1] + 1  # +1 to account for the first token that we ignored
        return first_eos



#!/usr/bin/env python3
# Copyright 2019 The Board of Trustees of the Leland Stanford Junior University
#
# Author: Giovanni Campagna <gcampagn@cs.stanford.edu>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#  list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#  this list of conditions and the following disclaimer in the documentation
#  and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#  contributors may be used to endorse or promote products derived from
#  this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import argparse

from . import arguments, cache_embeddings, calibrate, export, kfserver, predict, run_bootleg, server, train, write_kf_metrics
from .paraphrase import run_generation, run_lm_finetuning
from .paraphrase.scripts import clean_paraphrasing_dataset, dialog_to_tsv, split_dataset, transform_dataset
from .sts import sts_calculate_scores, sts_filter

subcommands = {
    # main commands
    'train': ('Train a model', arguments.parse_argv, train.main),
    'export': ('Export a trained model for serving', export.parse_argv, export.main),
    'predict': ('Evaluate a model, or compute predictions on a test dataset', predict.parse_argv, predict.main),
    'server': ('Export RPC interface to predict', server.parse_argv, server.main),
    'cache-embeddings': ('Download and cache embeddings', cache_embeddings.parse_argv, cache_embeddings.main),
    'train-paraphrase': ('Train a paraphraser model', run_lm_finetuning.parse_argv, run_lm_finetuning.main),
    'run-paraphrase': ('Run a paraphraser model', run_generation.parse_argv, run_generation.main),
    # calibration commands
    'calibrate': ('Train a confidence calibration model', calibrate.parse_argv, calibrate.main),
    # commands that work with datasets
    'transform-dataset': (
        'Apply transformations to a tab-separated dataset',
        transform_dataset.parse_argv,
        transform_dataset.main,
    ),
    'clean-paraphrasing-dataset': (
        'Select a clean subset from the ParaBank2 dataset',
        clean_paraphrasing_dataset.parse_argv,
        clean_paraphrasing_dataset.main,
    ),
    'dialog-to-tsv': (
        'Convert a dialog dataset to a turn-by-turn tab-separated format',
        dialog_to_tsv.parse_argv,
        dialog_to_tsv.main,
    ),
    'split-dataset': ('Split a dataset file into two files', split_dataset.parse_argv, split_dataset.main),
    # sts commands
    'calculate-paraphrase-sts': (
        'Calculate semantic similarity scores between a dataset and its paraphrase',
        sts_calculate_scores.parse_argv,
        sts_calculate_scores.main,
    ),
    'filter-paraphrase-sts': (
        'Filter paraphrases based on semantic similarity scores',
        sts_filter.parse_argv,
        sts_filter.main,
    ),
    # bootleg commands
    'bootleg-dump-features': (
        'Extract candidate features for named entity mentions in the dataset',
        run_bootleg.parse_argv,
        run_bootleg.main,
    ),
    # kf commands
    'kfserver': ('Export KFServing interface to predict', server.parse_argv, kfserver.main),
    'write-kf-metrics': ('Write KF evaluation metrics', write_kf_metrics.parse_argv, write_kf_metrics.main),
}


def main():
    parser = argparse.ArgumentParser(prog='genienlp')
    subparsers = parser.add_subparsers(dest='subcommand')
    for subcommand in subcommands:
        helpstr, get_parser, command_fn = subcommands[subcommand]
        get_parser(subparsers.add_parser(subcommand, help=helpstr))

    argv = parser.parse_args()
    subcommands[argv.subcommand][2](argv)


if __name__ == '__main__':
    main()



from typing import List

import torch
from transformers.models.gpt2.modeling_gpt2 import GPT2LMHeadModel


class GPT2Seq2Seq(GPT2LMHeadModel):
    def __init__(self, config):
        super().__init__(config)

    def set_token_ids(self, eos_token_id, sep_token_id, pad_token_id):
        self.eos_token_id = eos_token_id
        self.sep_token_id = sep_token_id
        self.pad_token_id = pad_token_id

    def pad_to_max_length(self, input_sequences: List[List[int]]):
        """
        Adds pad tokens to the left of each input_sequence
        """
        max_length = max([len(s) for s in input_sequences])
        copy_input_sequences = []
        for i in range(len(input_sequences)):
            copy_input_sequences.append([self.pad_token_id] * (max_length - len(input_sequences[i])) + input_sequences[i])

        return copy_input_sequences

    # TODO check if this function is used
    def enforce_repetition_penalty_(self, lprobs, batch_size, num_beams, prev_output_tokens, repetition_penalty):
        """repetition penalty from CTRL (https://arxiv.org/abs/1909.05858), but much faster on GPU"""
        if repetition_penalty == 1.0:
            return lprobs
        m = torch.scatter(input=torch.zeros_like(lprobs), dim=1, index=prev_output_tokens, value=1)
        m[: self.sep_token_id] = 0
        m[: self.pad_token_id] = 0
        # logger.info('m = ', m.shape)
        need_change = m * lprobs
        need_divide = need_change > 0
        need_multiply = need_change < 0
        lprobs = need_divide * lprobs / repetition_penalty + need_multiply * lprobs * repetition_penalty + (1 - m) * lprobs

    def generate(self, **kwargs):
        # change arguments so that they have the same meaning as seq2seq models
        if kwargs['bad_words_ids'] is None:
            kwargs['bad_words_ids'] = []
        kwargs['bad_words_ids'].append([self.sep_token_id])
        if kwargs['min_length'] is None:
            kwargs['min_length'] = 0
        kwargs['min_length'] += kwargs['input_ids'].shape[1]
        if kwargs['max_length'] is None:
            kwargs['max_length'] = 0
        kwargs['max_length'] += kwargs['input_ids'].shape[1]

        outputs = super().generate(**kwargs)
        sequences = outputs.sequences
        sequences = sequences[:, :].tolist()
        for i in range(len(outputs)):
            sequences[i] = [x for x in sequences[i] if x != self.pad_token_id]  # remove padding
            sequences[i] = sequences[i][
                sequences[i].index(self.sep_token_id) + 1 :
            ]  # only return the output (i.e. after sep_token)
        outputs.sequences = sequences

        return outputs

    def prepare_inputs_for_generation(self, input_ids, past=None, **kwargs):
        sep_token_position = (input_ids == self.sep_token_id).to(torch.long)
        assert (
            torch.sum(sep_token_position, dim=1) == 1
        ).all(), 'All input_ids must contain exactly one sep_token.' ' sep_token_position = %s\nsep_token_id = %d' % (
            str(sep_token_position),
            self.sep_token_id,
        )
        token_type_ids = torch.cumsum(sep_token_position, dim=1) - sep_token_position
        attention_mask = (input_ids != self.pad_token_id).to(torch.long)  # 0 means mask, 1 means no mask
        position_ids = (
            (torch.cumsum(attention_mask, dim=1) - 1) * (1 - token_type_ids)
            + (torch.cumsum(token_type_ids, dim=1) - 1) * token_type_ids
        ).clamp(min=0)
        token_type_ids = self.sep_token_id * (1 - token_type_ids) + self.eos_token_id * token_type_ids

        if past:
            input_ids = input_ids[:, -1].unsqueeze(-1)
            position_ids = position_ids[:, -1].unsqueeze(-1)
            token_type_ids = token_type_ids[:, -1].unsqueeze(-1)

        inputs = {
            "input_ids": input_ids,
            "position_ids": position_ids,
            "token_type_ids": token_type_ids,
            "attention_mask": attention_mask,
            "past_key_values": past,
        }
        return inputs
